import logging
import os
import os.path as op
import sys
from copy import copy, deepcopy

import cobra.flux_analysis
import cobra.manipulation
import numpy as np
import pandas as pd
from Bio import SeqIO
from cobra.core import DictList
from slugify import Slugify

import ssbio.core.modelpro
import ssbio.databases.ncbi
import ssbio.databases.patric
import ssbio.protein.sequence.properties.residues
import ssbio.protein.sequence.utils.alignment
import ssbio.protein.sequence.utils.blast
import ssbio.protein.sequence.utils.fasta
from ssbio import utils
from ssbio.core.object import Object
from ssbio.pipeline.gempro import GEMPRO

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
try:
    from IPython.display import clear_output
    have_ipython = True
    from tqdm import tqdm_notebook as tqdm
except ImportError:
    have_ipython = False
    from tqdm import tqdm

date = utils.Date()
custom_slugify = Slugify(safe_chars='-_.')
logging.basicConfig(stream=sys.stdout, level=logging.INFO)
log = logging.getLogger(__name__)


class ATLAS(Object):

    """Class to represent an ATLAS workflow to carry out multi-strain comparisons

    Main steps are:

    #. Strain-specific model construction based on orthologous genes & systems modeling
    #. Phylogenetic analysis to pick out important genes
    #. GEM-PRO of the "base strain"
    #. Structure property calculation & integrated structural systems analysis

    Each step may generate a report and also request additional files if something is missing

    """

    def __init__(self, atlas_name, root_dir, reference_gempro, reference_genome_path=None, description=None):
        """Prepare a GEM-PRO model for ATLAS analysis

        Args:
            atlas_name (str): Name of your ATLAS project
            root_dir (str): Path to where the folder named after ``atlas_name`` will be created.
            reference_gempro (GEMPRO): GEM-PRO model to use as the reference genome
            reference_genome_path (str): Path to reference genome FASTA file
            description (str): Optional string to describe your project

        """
        Object.__init__(self, id=atlas_name, description=description)

        # Create directories
        self._root_dir = None
        self.root_dir = root_dir

        self.strains = DictList()
        self.df_orthology_matrix = pd.DataFrame()
        # Mark if the orthology matrix has gene IDs (thus we need to retrieve seqs from the genome file) or if
        # it is in the orthology matrix itself
        self._orthology_matrix_has_sequences = False

        # Load the GEM-PRO (could be a model, could just be a list of genes)
        # Check if there is a genome file associated with this model - if not, write all sequences and use that
        self.reference_gempro = reference_gempro
        if not reference_genome_path and not self.reference_gempro.genome_path:
            self.reference_gempro.genome_path = self.reference_gempro.write_representative_sequences_file(outname=self.reference_gempro.id)
        else:
            self.reference_gempro.genome_path = reference_genome_path
            # TODO: must also check if reference_genome_path gene IDs can be matched to the reference_gempro

        # Also create an attribute
        self._empty_reference_gempro = None
        if self.reference_gempro.model:
            # If there is a SBML model associated with the GEMPRO, copy that model
            self._empty_reference_gempro = GEMPRO(gem_name='Copied reference GEM-PRO', gem=self.reference_gempro.model.copy())
            # Reset the GenePro attributes
            for x in self._empty_reference_gempro.genes:
                x.reset_protein()
        else:
            # Otherwise, just copy the list of genes over and rename the IDs
            strain_genes = [x.id for x in self.reference_gempro.genes]
            if len(strain_genes) == 0:
                raise ValueError('GEM-PRO has no genes, unable to run multi-strain analysis')
            self._empty_reference_gempro = GEMPRO(gem_name='Copied reference GEM-PRO', genes_list=strain_genes)

    @property
    def root_dir(self):
        """str: Directory where ATLAS project folder named after the attribute ``base_dir`` is located"""
        return self._root_dir

    @root_dir.setter
    def root_dir(self, path):
        if not path:
            raise ValueError('No path specified')

        if not op.exists(path):
            raise ValueError('{}: folder does not exist'.format(path))

        if self._root_dir:
            log.info('Changing root directory of project "{}" from {} to {}'.format(self.id, self.root_dir, path))

            if not op.exists(op.join(path, self.id)):
                raise IOError('Project "{}" does not exist in folder {}'.format(self.id, path))
        else:
            log.info('Creating project directory in folder {}'.format(path))

        self._root_dir = path

        for d in [self.base_dir, self.model_dir, self.data_dir,
                  self.sequences_dir, self.sequences_by_gene_dir, self.sequences_by_organism_dir]:
            ssbio.utils.make_dir(d)

        log.info('{}: project location'.format(self.base_dir))

    @property
    def base_dir(self):
        """str: ATLAS project folder"""
        if self.root_dir:
            return op.join(self.root_dir, self.id)
        else:
            return None

    @property
    def model_dir(self):
        """str: Directory where strain-specific GEMs are stored"""
        if self.base_dir:
            return op.join(self.base_dir, 'model')
        else:
            return None

    @property
    def data_dir(self):
        """str: Directory where all data (dataframes and more) will be stored"""
        if self.base_dir:
            return op.join(self.base_dir, 'data')
        else:
            return None

    @property
    def sequences_dir(self):
        """str: Base directory for genome protein sequences and alignments"""
        if self.base_dir:
            return op.join(self.base_dir, 'sequences')
        else:
            return None

    @property
    def sequences_by_gene_dir(self):
        """str: Directory where all gene specific information and pairwise alignments are stored"""
        if self.sequences_dir:
            return op.join(self.sequences_dir, 'by_gene')
        else:
            return None

    @property
    def sequences_by_organism_dir(self):
        """str: Directory where all strain specific genome and BLAST files are stored"""
        if self.sequences_dir:
            return op.join(self.sequences_dir, 'by_organism')
        else:
            return None

    # def _copy_reference_gempro(self, new_id):
    #     """Copy the base strain GEM-PRO into a new GEM-PRO with a specified ID.
    #
    #     Appends the model to the strains attribute.
    #
    #     Args:
    #         new_id (str): New ID to be assigned to the copied model
    #
    #     Returns:
    #         GEMPRO: copied GEM-PRO to represent the new strain
    #
    #     """
    #     logging.disable(logging.WARNING)
    #     if self.reference_gempro.model:
    #         # If there is a SBML model associated with the GEMPRO, copy that model
    #         copied_model = GEMPRO(gem_name=new_id, gem=self._model_to_copy.model.copy())
    #         copied_model.model.id = new_id
    #     else:
    #         # Otherwise, just copy the list of genes over and rename the IDs
    #         strain_genes = [x.id for x in self._model_to_copy.genes]
    #         copied_model = GEMPRO(gem_name=new_id, genes_list=strain_genes)
    #     # Re-enable logging
    #     logging.disable(logging.NOTSET)
    #
    #     self.strains.append(copied_model)
    #     log.debug('{}: new model copied from base model'.format(new_id))
    #
    #     return self.strains.get_by_id(new_id)

    def load_strain(self, strain_id, strain_genome_file):
        """Load a strain as a new GEM-PRO by its ID and associated genome file. Stored in the ``strains`` attribute.

        Args:
            strain_id (str): Strain ID
            strain_genome_file (str): Path to strain genome file

        """
        logging.disable(logging.WARNING)
        strain_gp = GEMPRO(gem_name=strain_id, genome_path=strain_genome_file)
        logging.disable(logging.NOTSET)

        self.strains.append(strain_gp)
        return self.strains.get_by_id(strain_id)

    def download_patric_genomes(self, ids, force_rerun=False):
        """Download genome files from PATRIC given a list of PATRIC genome IDs and load them as strains.

        Args:
            ids (str, list): PATRIC ID or list of PATRIC IDs
            force_rerun (bool): If genome files should be downloaded again even if they exist

        """
        ids = ssbio.utils.force_list(ids)

        counter = 0
        log.info('Downloading sequences from PATRIC...')
        for patric_id in tqdm(ids):
            f = ssbio.databases.patric.download_coding_sequences(patric_id=patric_id, seqtype='protein',
                                                                 outdir=self.sequences_by_organism_dir,
                                                                 force_rerun=force_rerun)
            if f:
                self.load_strain(patric_id, f)
                counter += 1
                log.debug('{}: downloaded sequence'.format(patric_id))
            else:
                log.warning('{}: unable to download sequence'.format(patric_id))

        log.info('Created {} new strain GEM-PROs, accessible at "strains" attribute'.format(counter))

    def get_orthology_matrix(self, pid_cutoff=None, bitscore_cutoff=None, evalue_cutoff=None, filter_condition='OR',
                             remove_strains_with_no_orthology=True,
                             remove_strains_with_no_differences=False,
                             remove_genes_not_in_base_model=True):
        """Create the orthology matrix by finding best bidirectional BLAST hits. Genes = rows, strains = columns

        Runs run_makeblastdb, run_bidirectional_blast, and calculate_bbh for protein sequences.

        Args:
            pid_cutoff (float): Minimum percent identity between BLAST hits to filter for in the range [0, 100]
            bitscore_cutoff (float): Minimum bitscore allowed between BLAST hits
            evalue_cutoff (float): Maximum E-value allowed between BLAST hits
            filter_condition (str): 'OR' or 'AND', how to combine cutoff filters. 'OR' gives more results since it
            is less stringent, as you will be filtering for hits with (>80% PID or >30 bitscore or <0.0001 evalue).
            remove_strains_with_no_orthology (bool): Remove strains which have no orthologous genes found
            remove_strains_with_no_differences (bool): Remove strains which have all the same genes as the base model.
                Default is False because since orthology is found using a PID cutoff, all genes may be present but
                differences may be on the sequence level.
            remove_genes_not_in_base_model (bool): Remove genes from the orthology matrix which are not present in our
                base model. This happens if we use a genome file for our model that has other genes in it.

        Returns:
            DataFrame: Orthology matrix calculated from best bidirectional BLAST hits.

        """
        # TODO: document and test other cutoffs

        # Get the path to the reference genome
        r_file = self.reference_gempro.genome_path

        bbh_files = {}

        log.info('Running bidirectional BLAST and finding best bidirectional hits (BBH)...')
        for strain_gempro in tqdm(self.strains):
            g_file = strain_gempro.genome_path

            # Run bidirectional BLAST
            log.debug('{} vs {}: Running bidirectional BLAST'.format(self.reference_gempro.id, strain_gempro.id))
            r_vs_g, g_vs_r = ssbio.protein.sequence.utils.blast.run_bidirectional_blast(reference=r_file,
                                                                                        other_genome=g_file,
                                                                                        dbtype='prot',
                                                                                        outdir=self.sequences_by_organism_dir)

            # Using the BLAST files, find the BBH
            log.debug('{} vs {}: Finding BBHs'.format(self.reference_gempro.id, strain_gempro.id))
            bbh = ssbio.protein.sequence.utils.blast.calculate_bbh(blast_results_1=r_vs_g, blast_results_2=g_vs_r,
                                                                   outdir=self.sequences_by_organism_dir)
            bbh_files[strain_gempro.id] = bbh

        # Make the orthologous genes matrix
        log.info('Creating orthology matrix from BBHs...')
        ortho_matrix = ssbio.protein.sequence.utils.blast.create_orthology_matrix(r_name=self.reference_gempro.id,
                                                                                  genome_to_bbh_files=bbh_files,
                                                                                  pid_cutoff=pid_cutoff,
                                                                                  bitscore_cutoff=bitscore_cutoff,
                                                                                  evalue_cutoff=evalue_cutoff,
                                                                                  filter_condition=filter_condition,
                                                                                  outname='{}_{}_orthology.csv'.format(self.reference_gempro.id, 'prot'),
                                                                                  outdir=self.data_dir)

        log.info('Saved orthology matrix at {}. See the "df_orthology_matrix" attribute.'.format(ortho_matrix))
        self.df_orthology_matrix = pd.read_csv(ortho_matrix, index_col=0)

        # Filter the matrix to genes only in our analysis, and also check for strains with no differences or no orthologous genes
        self._filter_orthology_matrix(remove_strains_with_no_orthology=remove_strains_with_no_orthology,
                                           remove_strains_with_no_differences=remove_strains_with_no_differences,
                                           remove_genes_not_in_base_model=remove_genes_not_in_base_model)

    # def load_manual_orthology_matrix(self, df, clean_names=True,
    #                                  remove_strains_with_no_orthology=True,
    #                                  remove_strains_with_no_differences=False,
    #                                  remove_genes_not_in_base_model=True):
    #     """Load a manually curated orthology matrix to use in ATLAS. Genes = rows, strains = columns.
    #
    #     Args:
    #         df (DataFrame): Pandas DataFrame with genes as the rows and strains as the columns
    #         clean_names (bool): Remove unwanted characters from gene names and strain IDs
    #         remove_strains_with_no_orthology (bool): Remove strains which have no orthologous genes found
    #         remove_strains_with_no_differences (bool): Remove strains which have all the same genes as the base model.
    #             Default is False because since orthology is found using a PID cutoff, all genes may be present but
    #             differences may be on the sequence level.
    #         remove_genes_not_in_base_model (bool): Remove genes from the orthology matrix which are not present in our
    #             base model. This happens if we use a genome file for our model that has other genes in it.
    #
    #     """
    #     self._orthology_matrix_has_sequences = True
    #
    #     if clean_names:
    #         new_rows = [custom_slugify(x) for x in df.index]
    #         new_cols = [custom_slugify(y) for y in df.columns]
    #         df.index = new_rows
    #         df.columns = new_cols
    #
    #     self.df_orthology_matrix = df
    #
    #     # Make the copies of the base model
    #     for strain_id in tqdm(self.df_orthology_matrix.columns):
    #         self._copy_reference_gempro(new_id=strain_id)
    #
    #     # Filter the strains and orthology matrix
    #     self._filter_orthology_matrix(remove_strains_with_no_orthology=remove_strains_with_no_orthology,
    #                                        remove_strains_with_no_differences=remove_strains_with_no_differences,
    #                                        remove_genes_not_in_base_model=remove_genes_not_in_base_model)

    def _filter_orthology_matrix(self,
                                 remove_strains_with_no_orthology=True,
                                 remove_strains_with_no_differences=False,
                                 remove_genes_not_in_base_model=True):
        """Filters the orthology matrix by removing genes not in our base model, and also
            removes strains from the analysis which have: 0 orthologous genes or no difference from the base strain.

        Args:
            remove_strains_with_no_orthology (bool): Remove strains which have no orthologous genes found
            remove_strains_with_no_differences (bool): Remove strains which have all the same genes as the base model.
                Default is False because since orthology is found using a PID cutoff, all genes may be present but
                differences may be on the sequence level.
            remove_genes_not_in_base_model (bool): Remove genes from the orthology matrix which are not present in our
                base model. This happens if we use a genome file for our model that has other genes in it.

        """

        if len(self.df_orthology_matrix) == 0:
            raise RuntimeError('Empty orthology matrix')

        initial_num_strains = len(self.strains)

        # Adding names to the row and column of the orthology matrix
        self.df_orthology_matrix = self.df_orthology_matrix.rename_axis('gene').rename_axis("strain", axis="columns")

        # Gene filtering (of the orthology matrix)
        if remove_genes_not_in_base_model:
            # Check for gene IDs that are in the model and not in the orthology matrix
            # This is probably because: the CDS FASTA file for the base strain did not contain the correct ID
            # for the gene and consequently was not included in the orthology matrix
            # Save these and report them
            reference_strain_gene_ids = [x.id for x in self.reference_gempro.genes]
            self.missing_in_orthology_matrix = [x for x in reference_strain_gene_ids if x not in self.df_orthology_matrix.index.tolist()]
            self.missing_in_reference_strain = [y for y in self.df_orthology_matrix.index.tolist() if y not in reference_strain_gene_ids]

            # Filter the matrix for genes within our base model only
            self.df_orthology_matrix = self.df_orthology_matrix[self.df_orthology_matrix.index.isin(reference_strain_gene_ids)]
            log.info('Filtered orthology matrix for genes present in base model')
            log.warning('{} genes are in your base model but not your orthology matrix, see the attribute "missing_in_orthology_matrix"'.format(len(self.missing_in_orthology_matrix)))
            log.warning('{} genes are in the orthology matrix but not your base model, see the attribute "missing_in_reference_strain"'.format(len(self.missing_in_reference_strain)))

        # Strain filtering
        for strain_gempro in self.strains.copy():
            if remove_strains_with_no_orthology:
                if strain_gempro.id not in self.df_orthology_matrix.columns:
                    self.strains.remove(strain_gempro)
                    log.info('{}: no orthologous genes found for this strain, removed from analysis.'.format(strain_gempro.id))
                    continue
                elif self.df_orthology_matrix[strain_gempro.id].isnull().all():
                    self.strains.remove(strain_gempro)
                    log.info('{}: no orthologous genes found for this strain, removed from analysis.'.format(strain_gempro.id))
                    continue

            if remove_strains_with_no_differences:
                not_in_strain = self.df_orthology_matrix[pd.isnull(self.df_orthology_matrix[strain_gempro.id])][strain_gempro.id].index.tolist()
                if len(not_in_strain) == 0:
                    self.strains.remove(strain_gempro)
                    log.info('{}: strain has no differences from the base, removed from analysis.')
                    continue

        log.info('{} strains to be analyzed, {} strains removed'.format(len(self.strains), initial_num_strains - len(self.strains)))

    def _pare_down_model(self, strain_gempro, genes_to_remove):
        """Mark genes as non-functional in a GEM-PRO. If there is a COBRApy model associated with it, the
            COBRApy method delete_model_genes is utilized to delete genes.

        Args:
            strain_gempro (GEMPRO): GEMPRO object
            genes_to_remove (list): List of gene IDs to remove from the model

        """
        # Filter out genes in genes_to_remove which do not show up in the model
        strain_genes = [x.id for x in strain_gempro.genes]
        genes_to_remove.extend(self.missing_in_orthology_matrix)
        genes_to_remove = list(set(genes_to_remove).intersection(set(strain_genes)))

        if len(genes_to_remove) == 0:
            log.info('{}: no genes marked non-functional'.format(strain_gempro.id))
            return
        else:
            log.debug('{}: {} genes to be marked non-functional'.format(strain_gempro.id, len(genes_to_remove)))

        # If a COBRApy model exists, utilize the delete_model_genes method
        if strain_gempro.model:
            strain_gempro.model._trimmed = False
            strain_gempro.model._trimmed_genes = []
            strain_gempro.model._trimmed_reactions = {}

            # Delete genes!
            cobra.manipulation.delete_model_genes(strain_gempro.model, genes_to_remove)

            if strain_gempro.model._trimmed:
                log.info('{}: marked {} genes as non-functional, '
                         'deactivating {} reactions'.format(strain_gempro.id, len(strain_gempro.model._trimmed_genes),
                                                            len(strain_gempro.model._trimmed_reactions)))
        # Otherwise, just mark the genes as non-functional
        else:
            for g in genes_to_remove:
                strain_gempro.genes.get_by_id(g).functional = False
            log.info('{}: marked {} genes as non-functional'.format(strain_gempro.id, len(genes_to_remove)))

    def _load_strain_sequences(self, strain_gempro):
        """Load strain sequences from the orthology matrix into the base model for comparisons, and into the
        strain-specific model itself.

        """
        if self._orthology_matrix_has_sequences:  # Load directly from the orthology matrix if it contains sequences
            strain_sequences = self.df_orthology_matrix[strain_gempro.id].to_dict()
        else:  # Otherwise load from the genome file if the orthology matrix contains gene IDs
            # Load the genome FASTA file
            log.debug('{}: loading strain genome CDS file'.format(strain_gempro.genome_path))
            strain_sequences = SeqIO.index(strain_gempro.genome_path, 'fasta')

        for strain_gene in strain_gempro.genes:
            if strain_gene.functional:
                if self._orthology_matrix_has_sequences:
                    strain_gene_key = strain_gene.id
                else:
                    # Pull the gene ID of the strain from the orthology matrix
                    strain_gene_key = self.df_orthology_matrix.loc[strain_gene.id, strain_gempro.id]
                    log.debug('{}: original gene ID to be pulled from strain fasta file'.format(strain_gene_key))

                # # Load into the base strain for comparisons
                ref_gene = self.reference_gempro.genes.get_by_id(strain_gene.id)
                new_id = '{}_{}'.format(strain_gene.id, strain_gempro.id)
                if ref_gene.protein.sequences.has_id(new_id):
                    log.debug('{}: sequence already loaded into reference model'.format(new_id))
                    continue
                ref_gene.protein.load_manual_sequence(seq=strain_sequences[strain_gene_key], ident=new_id,
                                                      set_as_representative=False)
                log.debug('{}: loaded sequence into reference model'.format(new_id))

                # Load into the strain GEM-PRO
                strain_gene.protein.load_manual_sequence(seq=strain_sequences[strain_gene_key], ident=new_id,
                                                         set_as_representative=True)
                log.debug('{}: loaded sequence into strain model'.format(new_id))

    def build_strain_specific_models(self, save_models=False):
        """Using the orthologous genes matrix, create and modify the strain specific models based on if orthologous
            genes exist.

        Also store the sequences directly in the reference GEM-PRO protein sequence attribute for the strains.
        """

        if len(self.df_orthology_matrix) == 0:
            raise RuntimeError('Empty orthology matrix')

        # Create an emptied copy of the reference GEM-PRO
        for strain_gempro in tqdm(self.strains):
            log.debug('{}: building strain specific model'.format(strain_gempro.id))

            # For each genome, load the metabolic model or genes from the reference GEM-PRO
            logging.disable(logging.WARNING)
            if self._empty_reference_gempro.model:
                strain_gempro.load_cobra_model(self._empty_reference_gempro.model)
            elif self._empty_reference_gempro.genes:
                strain_gempro.genes = [x.id for x in self._empty_reference_gempro.genes]
            logging.disable(logging.NOTSET)

            # Get a list of genes which do not have orthology in the strain
            not_in_strain = self.df_orthology_matrix[pd.isnull(self.df_orthology_matrix[strain_gempro.id])][strain_gempro.id].index.tolist()

            # Mark genes non-functional
            self._pare_down_model(strain_gempro=strain_gempro, genes_to_remove=not_in_strain)

            # Load sequences into the base and strain models
            self._load_strain_sequences(strain_gempro=strain_gempro)

            if save_models:
                cobra.io.save_json_model(model=strain_gempro.model,
                                         filename=op.join(self.model_dir, '{}.json'.format(strain_gempro.id)))
                strain_gempro.save_pickle(op.join(self.model_dir, '{}_gp.pckl'.format(strain_gempro.id)))


        log.info('Created {} new strain-specific models and loaded in sequences'.format(len(self.strains)))

    def align_orthologous_genes_pairwise(self, gapopen=10, gapextend=0.5):
        """For each gene in the base strain, run a pairwise alignment for all orthologous gene sequences to it."""
        for ref_gene in tqdm(self.reference_gempro.genes):
            if len(ref_gene.protein.sequences) > 1:
                alignment_dir = op.join(self.sequences_by_gene_dir, ref_gene.id)
                if not op.exists(alignment_dir):
                    os.mkdir(alignment_dir)
                ref_gene.protein.pairwise_align_sequences_to_representative(gapopen=gapopen, gapextend=gapextend,
                                                                            outdir=alignment_dir, parse=True)

    def align_orthologous_genes_multiple(self):
        """For each gene in the base strain, run a multiple alignment to all orthologous strain genes"""
        pass

    def get_atlas_summary_df(self):
        """Create a single data frame which summarizes all genes per row.

        Returns:
            DataFrame: Pandas DataFrame of the results

        """
        all_info = []
        for g in self.reference_gempro.genes_with_a_representative_sequence:
            info = {}
            info['Gene_ID'] = g.id
            info['Gene_name'] = g.name

            # Protein object
            p = g.protein
            info['Protein_sequences'] = len(p.sequences)
            info['Protein_structures'] = len(p.structures)

            # SeqProp
            rseq = p.representative_sequence
            info['RepSeq_ID'] = rseq.id
            info['RepSeq_sequence_length'] = rseq.seq_len
            info['RepSeq_num_sequence_alignments'] = len([x for x in p.sequence_alignments if x.annotations['ssbio_type'] == 'seqalign'])
            info['RepSeq_num_structure_alignments'] = len([x for x in p.sequence_alignments if x.annotations['ssbio_type'] == 'structalign'])

            # SeqRecord annotations (properties calculated that summarize the whole sequence)
            for annotation_name, annotation in rseq.annotations.items():
                info['RepSeq_' + annotation_name] = annotation

            # SeqRecord alignment annotations
            all_num_mutations = []
            all_num_deletions = []
            all_len_deletions = []
            all_num_insertions = []
            all_len_insertions = []
            all_percent_identity = []
            all_percent_similarity = []
            for aln in p.sequence_alignments:
                # Gather the strain speicific stuff
                if '{}_'.format(p.id) not in aln.annotations['b_seq']:
                    continue
                info[aln.annotations['b_seq'].split('{}_'.format(p.id))[1]] = aln.annotations['percent_identity']

                # Gather the percent identities/similarities
                all_percent_identity.append(aln.annotations['percent_identity'])
                all_percent_similarity.append(aln.annotations['percent_similarity'])

                # Gather the number of residues that are mutated (filter for different mutations of same residue)
                num_mutations = len(list(set([x[1] for x in aln.annotations['mutations']])))
                all_num_mutations.append(num_mutations)

                # Gather the number of deletions as well as the length of the deletion
                if not aln.annotations['deletions']:
                    num_deletions = 0
                    len_deletions = [0]
                else:
                    num_deletions = len(aln.annotations['deletions'])
                    len_deletions = [x[1] for x in aln.annotations['deletions']]
                all_num_deletions.append(num_deletions)
                # Get the total length of the deletion for this one strain
                avg_len_deletions = np.sum(len_deletions)
                all_len_deletions.append(avg_len_deletions)

                # Gather the number of insertions as well as the length of the insertion
                if not aln.annotations['insertions']:
                    num_insertions = 0
                    len_insertions = [0]
                else:
                    num_insertions = len(aln.annotations['insertions'])
                    len_insertions = [x[1] for x in aln.annotations['insertions']]
                all_num_insertions.append(num_insertions)
                # Get the total length of insertion for this one strain
                avg_len_insertions = np.sum(len_insertions)
                all_len_insertions.append(avg_len_insertions)

            info['ATLAS_mean_num_mutations'] = np.mean(all_num_mutations)
            info['ATLAS_mean_num_deletions'] = np.mean(all_num_deletions)
            info['ATLAS_mean_len_deletions'] = np.mean(all_len_deletions)
            info['ATLAS_mean_num_insertions'] = np.mean(all_num_insertions)
            info['ATLAS_mean_len_insertions'] = np.mean(all_len_insertions)
            info['ATLAS_mean_percent_identity'] = np.mean(all_percent_identity)
            info['ATLAS_mean_percent_similarity'] = np.mean(all_percent_similarity)

            # Other mutation analysis
            single, fingerprint = p.sequence_mutation_summary()

            # Mutations that show up in more than 10% of strains
            singles = []
            for k, v in single.items():
                k = [str(x) for x in k]
                if len(v) / len(p.sequence_alignments) >= 0.01:
                    singles.append(''.join(k))  # len(v) is the number of strains which have this mutation
            info['ATLAS_popular_mutations'] = ';'.join(singles)

            # Mutation groups that show up in more than 10% of strains
            allfingerprints = []
            for k, v in fingerprint.items():
                if len(v) / len(p.sequence_alignments) >= 0.01:
                    fingerprints = []
                    for m in k:
                        y = [str(x) for x in m]
                        fingerprints.append(''.join(y))
                    allfingerprints.append('-'.join(fingerprints))
            info['ATLAS_popular_mutation_groups'] = ';'.join(allfingerprints)

            # StructProp
            rstruct = p.representative_structure
            if rstruct:
                if rstruct.structure_file:
                    info['RepStruct_ID'] = rstruct.id
                    info['RepStruct_is_experimental'] = rstruct.is_experimental
                    info['RepStruct_description'] = rstruct.description
                    info['RepStruct_repseq_coverage'] = p.representative_chain_seq_coverage

                    # ChainProp
                    rchain = p.representative_chain
                    info['RepChain_ID'] = rchain

                    # ChainProp SeqRecord annotations
                    rchain_sr = rstruct.chains.get_by_id(rchain).seq_record
                    for annotation_name, annotation in rchain_sr.annotations.items():
                        info['RepChain_' + annotation_name] = annotation

            all_info.append(info)

        cols = ['Gene_ID', 'Gene_name', 'Protein_sequences', 'Protein_structures',
                'RepSeq_ID', 'RepSeq_sequence_length',
                'RepSeq_num_sequence_alignments', 'RepSeq_num_structure_alignments',
                'RepStruct_ID', 'RepChain_ID', 'RepStruct_description',
                'RepStruct_is_experimental', 'RepStruct_repseq_coverage',
                'ATLAS_mean_percent_identity', 'ATLAS_mean_percent_similarity', 'ATLAS_mean_num_mutations',
                'ATLAS_popular_mutations', 'ATLAS_popular_mutation_groups', 'ATLAS_mean_num_deletions',
                'ATLAS_mean_num_insertions', 'ATLAS_mean_len_deletions', 'ATLAS_mean_len_insertions',
                'RepSeq_aromaticity', 'RepSeq_instability_index', 'RepSeq_isoelectric_point', 'RepSeq_molecular_weight',
                'RepSeq_monoisotopic', 'RepSeq_num_tm_helix-tmhmm', 'RepSeq_percent_acidic', 'RepSeq_percent_aliphatic',
                'RepSeq_percent_aromatic', 'RepSeq_percent_B-sspro8', 'RepSeq_percent_basic',
                'RepSeq_percent_buried-accpro', 'RepSeq_percent_buried-accpro20', 'RepSeq_percent_C-sspro',
                'RepSeq_percent_C-sspro8', 'RepSeq_percent_charged', 'RepSeq_percent_E-sspro',
                'RepSeq_percent_E-sspro8', 'RepSeq_percent_exposed-accpro', 'RepSeq_percent_exposed-accpro20',
                'RepSeq_percent_G-sspro8', 'RepSeq_percent_H-sspro', 'RepSeq_percent_H-sspro8',
                'RepSeq_percent_helix_naive', 'RepSeq_percent_I-sspro8', 'RepSeq_percent_non-polar',
                'RepSeq_percent_polar', 'RepSeq_percent_S-sspro8', 'RepSeq_percent_small',
                'RepSeq_percent_strand_naive', 'RepSeq_percent_T-sspro8', 'RepSeq_percent_tiny',
                'RepSeq_percent_turn_naive', 'RepChain_percent_B-dssp', 'RepChain_percent_C-dssp',
                'RepChain_percent_E-dssp', 'RepChain_percent_G-dssp', 'RepChain_percent_H-dssp',
                'RepChain_percent_I-dssp', 'RepChain_percent_S-dssp', 'RepChain_percent_T-dssp',
                'RepChain_SSBOND-biopython']
        cols.extend([x.id for x in self.strains])

        df_atlas_summary = pd.DataFrame(all_info, columns=cols)
        # Drop columns that don't have anything in them
        df_atlas_summary.dropna(axis=1, how='all', inplace=True)

        return df_atlas_summary

    def get_atlas_per_gene_mutation_df(self, gene_id):
        """Create a single data frame which summarizes a gene and its mutations.

        Args:
            gene_id (str): Gene ID in the base model

        Returns:
            DataFrame: Pandas DataFrame of the results

        """
        # TODO: also count: number of unique mutations (have to consider position, amino acid change)
        # TODO: keep track of strain with most mutations, least mutations
        # TODO: keep track of strains that conserve the length of the protein, others that extend or truncate it
        # need statistical test for that too (how long is "extended"/"truncated"?)
        # TODO: number of strains with at least 1 mutations
        # TODO: number of strains with <5% mutated, 5-10%, etc

        g = self.reference_gempro.genes.get_by_id(gene_id)

        single, fingerprint = g.protein.sequence_mutation_summary(alignment_type='seqalign')

        structure_type_suffix = 'NA'
        appender = []

        for k, strains in single.items():
            # Mutations in the strain
            to_append = {}
            orig_res = k[0]
            resnum = int(k[1])
            mutated_res = k[2]
            num_strains_mutated = len(strains)
            strain_ids = [str(x.split(g.id + '_')[1]) for x in strains]
            to_append['ref_residue'] = orig_res
            to_append['ref_resnum'] = resnum
            to_append['strain_residue'] = mutated_res
            to_append['num_strains_mutated'] = num_strains_mutated
            to_append['strains_mutated'] = ';'.join(strain_ids)
            to_append['at_disulfide_bridge'] = False

            # Residue properties
            origres_props = ssbio.protein.sequence.properties.residues.residue_biochemical_definition(orig_res)
            mutres_props = ssbio.protein.sequence.properties.residues.residue_biochemical_definition(mutated_res)
            to_append['ref_residue_prop'] = origres_props
            to_append['strain_residue_prop'] = mutres_props

            # Grantham score - score a mutation based on biochemical properties
            grantham_s, grantham_txt = ssbio.protein.sequence.properties.residues.grantham_score(orig_res, mutated_res)
            to_append['grantham_score'] = grantham_s
            to_append['grantham_annotation'] = grantham_txt

            # Get all per residue annotations - predicted from sequence and calculated from structure
            to_append.update(g.protein.get_residue_annotations(seq_resnum=resnum, use_representatives=True))

            # Check structure type
            if g.protein.representative_structure:
                if g.protein.representative_structure.is_experimental:
                    to_append['structure_type'] = 'EXP'
                else:
                    to_append['structure_type'] = 'HOM'

                # At disulfide bond?
                repchain = g.protein.representative_chain
                repchain_annotations = g.protein.representative_structure.chains.get_by_id(repchain).seq_record.annotations
                if 'SSBOND-biopython' in repchain_annotations:
                    structure_resnum = g.protein.map_seqprop_resnums_to_structprop_resnums(resnums=resnum,
                                                                                           use_representatives=True)
                    if resnum in structure_resnum:
                        ssbonds = repchain_annotations['SSBOND-biopython']
                        ssbonds_res = []
                        for x in ssbonds:
                            ssbonds_res.append(x[0])
                            ssbonds_res.append(x[1])

                        if structure_resnum in ssbonds_res:
                            to_append['at_disulfide_bridge'] = True

            appender.append(to_append)

        if not appender:
            return pd.DataFrame()

        cols = ['ref_residue', 'ref_resnum', 'strain_residue', 'num_strains_mutated', 'strains_mutated',
                'ref_residue_prop', 'strain_residue_prop', 'grantham_score', 'grantham_annotation',
                'at_disulfide_bridge',
                'seq_SS-sspro', 'seq_SS-sspro8', 'seq_RSA-accpro', 'seq_RSA-accpro20', 'seq_TM-tmhmm',
                'struct_SS-dssp', 'struct_RSA-dssp', 'struct_ASA-dssp',
                'struct_CA_DEPTH-msms', 'struct_RES_DEPTH-msms',
                'struct_PHI-dssp', 'struct_PSI-dssp',
                'struct_resnum', 'struct_residue'
                'strains_mutated']

        df_gene_summary = pd.DataFrame.from_records(appender, columns=cols)

        # Drop columns that don't have anything in them
        df_gene_summary.dropna(axis=1, how='all', inplace=True)

        df_gene_summary.sort_values(by='ref_resnum', inplace=True)
        df_gene_summary = df_gene_summary.set_index('ref_resnum')
        return df_gene_summary

    def download_mutation_images(self):
        # TODO: dunno if this works
        import ipywidgets
        import math

        views = []
        for g in self.reference_gempro.genes:
            if g.protein.representative_structure:
                view = g.protein.view_all_mutations(alignment_type='seqalign', grouped=False, structure_opacity=0.5,
                                                    opacity_range=(0.6, 1), scale_range=(.5, 5))
                view._remote_call("setSize", target='Widget', args=['300px', '300px'])
                view.download_image(filename='{}_{}_mutations.png'.format(g.id, g.name))
                views.append(view)

        hboxes = [ipywidgets.HBox(views[i * 3:i * 3 + 3])
                  for i in range(int(math.ceil(len(views) / 3.0)))]
        vbox = ipywidgets.VBox(hboxes)
        return vbox


import os
import pandas as pd
import os.path as op
import logging
from ssbio.core.object import Object
import ssbio.utils

log = logging.getLogger(__name__)


class DOCK(Object):

    """Class to prepare a structure file for docking with DOCK6.

    Attributes:

    """

    def __init__(self, structure_id, pdb_file, amb_file, flex1_file, flex2_file, root_dir=None):
        """Initialize a DOCK6 project.

        Args:

        """

        super(DOCK, self).__init__(id=structure_id, description='DOCK6 preparation')
        self._root_dir = None
        self.structure_path = pdb_file

        if root_dir:
            self.root_dir = root_dir
        else:
            self.root_dir = self.structure_dir

        self.dockprep_path = None
        self.receptormol2_path = None
        self.receptorpdb_path = None
        self.dms_path = None
        self.sphgen_path = None
        self.bindingsite_path = None
        self.sphsel_path = None
        self.box_path = None
        self.grid_path = None

        self.dock_flexible_outfile = None
        self.dock_flexible_scored_result = None
        self.dock_flexible_conformers_result = None

        self.amb_file = amb_file
        self.flex1_file = flex1_file
        self.flex2_file = flex2_file

        log.debug('{}: created DOCK6 project folder at {}'.format(structure_id, self.dock_dir))

    @property
    def root_dir(self):
        """str: Directory where DOCK project folder is located"""
        return self._root_dir

    @root_dir.setter
    def root_dir(self, path):
        if not path:
            raise ValueError('No path specified')

        if not op.exists(path):
            raise ValueError('{}: folder does not exist'.format(path))

        if self._root_dir:
            log.debug('Changing root directory of DOCK project for "{}" from {} to {}'.format(self.id, self.root_dir, path))

            if not op.exists(op.join(path, self.id)):
                raise IOError('DOCK project "{}" does not exist in folder {}'.format(self.id, path))

        self._root_dir = path

        for d in [self.dock_dir]:
            ssbio.utils.make_dir(d)

    @property
    def dock_dir(self):
        """str: DOCK folder"""
        if self.root_dir:
            return op.join(self.root_dir, self.id + '_DOCK')
        else:
            log.warning('Root directory not set')
            return None

    @property
    def structure_dir(self):
        if not self._structure_dir:
            raise OSError('No structure folder set')
        return self._structure_dir

    @structure_dir.setter
    def structure_dir(self, path):
        if path and not op.exists(path):
            raise OSError('{}: folder does not exist'.format(path))

        self._structure_dir = path

    @property
    def structure_path(self):
        if not self.structure_file:
            raise OSError('Metadata file not loaded')

        path = op.join(self.structure_dir, self.structure_file)
        if not op.exists(path):
            raise OSError('{}: file does not exist'.format(path))
        return path

    @structure_path.setter
    def structure_path(self, path):
        """Provide pointers to the paths of the structure file

        Args:
            path: Path to structure file

        """
        if not path:
            self.structure_dir = None
            self.structure_file = None

        else:
            if not op.exists(path):
                raise OSError('{}: file does not exist!'.format(path))

            if not op.dirname(path):
                self.structure_dir = '.'
            else:
                self.structure_dir = op.dirname(path)
            self.structure_file = op.basename(path)

    def dockprep(self, force_rerun=False):
        """Prepare a PDB file for docking by first converting it to mol2 format.

        Args:
            force_rerun (bool): If method should be rerun even if output file exists

        """
        log.debug('{}: running dock preparation...'.format(self.id))

        prep_mol2 = op.join(self.dock_dir, '{}_prep.mol2'.format(self.id))
        prep_py = op.join(self.dock_dir, "prep.py")

        if ssbio.utils.force_rerun(flag=force_rerun, outfile=prep_mol2):
            with open(prep_py, "w") as f:
                f.write('import chimera\n')
                f.write('from DockPrep import prep\n')
                f.write('models = chimera.openModels.list(modelTypes=[chimera.Molecule])\n')
                f.write('prep(models)\n')
                f.write('from WriteMol2 import writeMol2\n')
                f.write('writeMol2(models, "{}")\n'.format(prep_mol2))

            cmd = 'chimera --nogui {} {}'.format(self.structure_path, prep_py)
            os.system(cmd)
            os.remove(prep_py)
            os.remove('{}c'.format(prep_py))

        if ssbio.utils.is_non_zero_file(prep_mol2):
            self.dockprep_path = prep_mol2
            log.debug('{}: successful dockprep execution'.format(self.dockprep_path))
        else:
            log.critical('{}: dockprep failed to run on PDB file'.format(self.structure_path))

    def protein_only_and_noH(self, keep_ligands=None, force_rerun=False):
        """Isolate the receptor by stripping everything except protein and specified ligands.

        Args:
            keep_ligands (str, list): Ligand(s) to keep in PDB file
            force_rerun (bool): If method should be rerun even if output file exists

        """
        log.debug('{}: running protein receptor isolation...'.format(self.id))

        if not self.dockprep_path:
            return ValueError('Please run dockprep')

        receptor_mol2 = op.join(self.dock_dir, '{}_receptor.mol2'.format(self.id))
        receptor_noh = op.join(self.dock_dir, '{}_receptor_noH.pdb'.format(self.id))

        prly_com = op.join(self.dock_dir, "prly.com")

        if ssbio.utils.force_rerun(flag=force_rerun, outfile=receptor_noh):
            with open(prly_com, "w") as f:
                f.write('open {}\n'.format(self.dockprep_path))

                keep_str = 'delete ~protein'
                if keep_ligands:
                    keep_ligands = ssbio.utils.force_list(keep_ligands)
                    for res in keep_ligands:
                        keep_str += ' & ~:{} '.format(res)
                keep_str = keep_str.strip() + '\n'
                f.write(keep_str)

                f.write('write format mol2 0 {}\n'.format(receptor_mol2))
                f.write('delete element.H\n')
                f.write('write format pdb 0 {}\n'.format(receptor_noh))

            cmd = 'chimera --nogui {}'.format(prly_com)
            os.system(cmd)
            os.remove(prly_com)

        if ssbio.utils.is_non_zero_file(receptor_mol2) and ssbio.utils.is_non_zero_file(receptor_noh):
            self.receptormol2_path = receptor_mol2
            self.receptorpdb_path = receptor_noh
            log.debug('{}: successful receptor isolation (mol2)'.format(self.receptormol2_path))
            log.debug('{}: successful receptor isolation (pdb)'.format(self.receptorpdb_path))
        else:
            log.critical('{}: protein_only_and_noH failed to run on dockprep file'.format(self.dockprep_path))

    def dms_maker(self, force_rerun=False):
        """Create surface representation (dms file) of receptor

        Args:
            force_rerun (bool): If method should be rerun even if output file exists

        """
        log.debug('{}: running surface representation maker...'.format(self.id))

        if not self.receptorpdb_path:
            return ValueError('Please run protein_only_and_noH')

        dms = op.join(self.dock_dir, '{}_receptor.dms'.format(self.id))

        if ssbio.utils.force_rerun(flag=force_rerun, outfile=dms):
            cmd = 'dms {} -n -w 1.4 -o {}'.format(self.receptorpdb_path, dms)
            os.system(cmd)

        self.dms_path = dms

        if ssbio.utils.is_non_zero_file(dms):
            self.dms_path = dms
            log.debug('{}: successful dms execution'.format(self.dms_path))
        else:
            log.critical('{}: dms_maker failed to run on receptor file'.format(self.receptorpdb_path))

    def sphgen(self, force_rerun=False):
        """Create sphere representation (sph file) of receptor from the surface representation

        Args:
            force_rerun (bool): If method should be rerun even if output file exists

        """
        log.debug('{}: running sphere generation...'.format(self.id))

        if not self.dms_path:
            return ValueError('Please run dms_maker')

        sph = op.join(self.dock_dir, '{}_receptor.sph'.format(self.id))
        insph = op.join(self.dock_dir, 'INSPH')

        if ssbio.utils.force_rerun(flag=force_rerun, outfile=sph):
            with open(insph, "w") as f:
                f.write("{}\n".format(self.dms_path))
                f.write("R\n")
                f.write("X\n")
                f.write("0.0\n")
                f.write("4.0\n")
                f.write("1.4\n")
                f.write("{}\n".format(sph))

            os.chdir(self.dock_dir)
            cmd = "sphgen_cpp"
            os.system(cmd)
            os.remove(insph)

        if ssbio.utils.is_non_zero_file(sph):
            self.sphgen_path = sph
            log.debug('{}: successful sphgen execution'.format(self.sphgen_path))
        else:
            log.critical('{}: sphgen_cpp failed to run on dms file'.format(self.dms_path))

    def binding_site_mol2(self, residues, force_rerun=False):
        """Create mol2 of only binding site residues from the receptor

        This function will take in a .pdb file (preferably the _receptor_noH.pdb file)
        and a string of residues (eg: '144,170,199') and delete all other residues in the
        .pdb file. It then saves the coordinates of the selected residues as a .mol2 file.
        This is necessary for Chimera to select spheres within the radius of the binding
        site.

        Args:
            residues (str): Comma separated string of residues (eg: '144,170,199')
            force_rerun (bool): If method should be rerun even if output file exists

        """
        log.debug('{}: running binding site isolation...'.format(self.id))

        if not self.receptorpdb_path:
            return ValueError('Please run protein_only_and_noH')

        prefix = self.id + '_' + 'binding_residues'
        mol2maker = op.join(self.dock_dir, '{}_make_mol2.py'.format(prefix))
        outfile = op.join(self.dock_dir, '{}.mol2'.format(prefix))

        if ssbio.utils.force_rerun(flag=force_rerun, outfile=outfile):
            with open(mol2maker, 'w') as mol2_maker:
                mol2_maker.write('#! /usr/bin/env python\n')
                mol2_maker.write('from chimera import runCommand\n')
                mol2_maker.write('runCommand("open {}")\n'.format(self.receptorpdb_path))
                mol2_maker.write('runCommand("delete ~:{}")\n'.format(residues))
                mol2_maker.write('runCommand("write format mol2 resnum 0 {}")\n'.format(outfile))
                mol2_maker.write('runCommand("close all")')

            cmd = 'chimera --nogui {}'.format(mol2maker)
            os.system(cmd)
            os.remove(mol2maker)
            os.remove('{}c'.format(mol2maker))

        if ssbio.utils.is_non_zero_file(outfile):
            self.bindingsite_path = outfile
            log.debug('{}: successful binding site isolation'.format(self.bindingsite_path))
        else:
            log.critical('{}: binding_site_mol2 failed to run on receptor file'.format(self.receptorpdb_path))

    def sphere_selector_using_residues(self, radius, force_rerun=False):
        """Select spheres based on binding site residues

        Args:
            radius (int, float): Radius around binding residues to dock to
            force_rerun (bool): If method should be rerun even if output file exists

        """
        log.debug('{}: running sphere selector...'.format(self.id))

        if not self.sphgen_path or not self.bindingsite_path:
            return ValueError('Please run sphgen and binding_site_mol2')

        selsph = op.join(self.dock_dir, '{}_selsph_binding.sph'.format(self.id))

        if ssbio.utils.force_rerun(flag=force_rerun, outfile=selsph):
            cmd = "sphere_selector {} {} {}".format(self.sphgen_path, self.bindingsite_path, radius)
            rename = "mv selected_spheres.sph {}".format(selsph)

            os.system(cmd)
            os.system(rename)

        if ssbio.utils.is_non_zero_file(selsph):
            self.sphsel_path = selsph
            log.debug('{}: successful sphere selection'.format(self.sphsel_path))
        else:
            log.critical('{}: sphere_selector_using_residues failed to run on sph file'.format(self.sphgen_path))

    # def split_sph(self, force_rerun=False):
    #     """TODO: documentation? what was this used for"""
    #
    #     selsph = op.join(self.dock_dir, '{}_selsph.sph'.format(self.id))
    #
    #     if ssbio.utils.force_rerun(flag=force_rerun, outfile=selsph):
    #         with open(self.sphgen_path, "r") as f:
    #             text = f.read()
    #             f.seek(0)
    #             lines = f.readlines()
    #             paragraphs = re.split("cluster    ...  number of spheres in cluster   ...\n", text)
    #
    #             with open(selsph, "w") as f2:
    #                 f2.write(lines[1])
    #                 f2.write(paragraphs[1])
    #
    #     return selsph

    def showbox(self, force_rerun=False):
        """Create the dummy PDB box around the selected spheres.

        Args:
            force_rerun (bool): If method should be rerun even if output file exists

        """
        log.debug('{}: running box maker...'.format(self.id))

        if not self.sphsel_path:
            return ValueError('Please run sphere_selector_using_residues')

        boxfile = op.join(self.dock_dir, "{}_box.pdb".format(self.id))
        boxscript = op.join(self.dock_dir, "{}_box.in".format(self.id))

        if ssbio.utils.force_rerun(flag=force_rerun, outfile=boxfile):
            with open(boxscript, "w") as f:
                f.write("Y\n")
                f.write("0\n")
                f.write("{}\n".format(op.basename(self.sphsel_path)))
                f.write("1\n")
                f.write("{}".format(op.basename(boxfile)))

            cmd = "showbox < {}".format(boxscript)
            os.chdir(self.dock_dir)
            os.system(cmd)

        if ssbio.utils.is_non_zero_file(boxfile):
            self.box_path = boxfile
            log.debug('{}: successful box creation'.format(self.box_path))
        else:
            log.critical('{}: showbox failed to run on selected spheres file'.format(self.sphsel_path))

    def grid(self, force_rerun=False):
        """Create the scoring grid within the dummy box.

        Args:
            force_rerun (bool): If method should be rerun even if output file exists

        """
        log.debug('{}: running grid maker...'.format(self.id))

        if not self.receptormol2_path or not self.box_path:
            return ValueError('Please run protein_only_and_noH and showbox')

        gridscript = op.join(self.dock_dir, "{}_grid.in".format(self.id))
        out_name = op.join(self.dock_dir, "{}_grid.out".format(self.id))

        if ssbio.utils.force_rerun(flag=force_rerun, outfile=out_name):
            with open(gridscript, "w") as f:
                grid_text = """compute_grids                  yes
grid_spacing                   0.3
output_molecule                no
contact_score                  yes
contact_cutoff_distance        4.5

energy_score                   yes
energy_cutoff_distance         9999

atom_model                     all
attractive_exponent            6
repulsive_exponent             12
distance_dielectric            yes
dielectric_factor              4

bump_filter                    yes
bump_overlap                   0.75

receptor_file                  {}
box_file                       {}
vdw_definition_file            {}
score_grid_prefix              {}_grid
            """.format(op.basename(self.receptormol2_path), op.basename(self.box_path), self.amb_file, self.id)

                f.write(grid_text)

            os.chdir(self.dock_dir)
            cmd = "grid -i {} -o {}".format(op.basename(gridscript), op.basename(out_name))
            os.system(cmd)

        if ssbio.utils.is_non_zero_file(out_name):
            self.grid_path = out_name
            log.debug('{}: successful grid creation'.format(self.grid_path))
        else:
            log.critical('{}: grid failed to run on receptor + box file'.format(self.box_path))

    def do_dock6_flexible(self, ligand_path, force_rerun=False):
        """Dock a ligand to the protein.

        Args:
            ligand_path (str): Path to ligand (mol2 format) to dock to protein
            force_rerun (bool): If method should be rerun even if output file exists

        """
        log.debug('{}: running DOCK6...'.format(self.id))

        ligand_name = os.path.basename(ligand_path).split('.')[0]
        in_name = op.join(self.dock_dir, "{}_{}_flexdock.in".format(self.id, ligand_name))
        out_name = op.join(self.dock_dir, "{}_{}_flexdock.out".format(self.id, ligand_name))

        conformers_out = op.join(self.dock_dir, '{}_{}_flexdock_conformers.mol2'.format(self.id, ligand_name))
        scored_out = op.join(self.dock_dir, '{}_{}_flexdock_scored.mol2'.format(self.id, ligand_name))
        ranked_out = op.join(self.dock_dir, '{}_{}_flexdock_ranked.mol2'.format(self.id, ligand_name))

        if ssbio.utils.force_rerun(flag=force_rerun, outfile=ranked_out):
            with open(in_name, "w") as f:
                dock_text = """ligand_atom_file                                             {}
limit_max_ligands                                            no
skip_molecule                                                no
read_mol_solvation                                           no
calculate_rmsd                                               no
use_database_filter                                          no
orient_ligand                                                yes
automated_matching                                           yes
receptor_site_file                                           {}
max_orientations                                             500
critical_points                                              no
chemical_matching                                            no
use_ligand_spheres                                           no
use_internal_energy                                          yes
internal_energy_rep_exp                                      12
flexible_ligand                                              yes
user_specified_anchor                                        no
limit_max_anchors                                            no
min_anchor_size                                              5
pruning_use_clustering                                       yes
pruning_max_orients                                          100
pruning_clustering_cutoff                                    100
pruning_conformer_score_cutoff                               100
use_clash_overlap                                            no
write_growth_tree                                            no
bump_filter                                                  yes
bump_grid_prefix                                             {}
score_molecules                                              yes
contact_score_primary                                        no
contact_score_secondary                                      no
grid_score_primary                                           yes
grid_score_secondary                                         no
grid_score_rep_rad_scale                                     1
grid_score_vdw_scale                                         1
grid_score_es_scale                                          1
grid_score_grid_prefix                                       {}
multigrid_score_secondary                                    no
dock3.5_score_secondary                                      no
continuous_score_secondary                                   no
descriptor_score_secondary                                   no
gbsa_zou_score_secondary                                     no
gbsa_hawkins_score_secondary                                 no
SASA_descriptor_score_secondary                              no
amber_score_secondary                                        no
minimize_ligand                                              yes
minimize_anchor                                              yes
minimize_flexible_growth                                     yes
use_advanced_simplex_parameters                              no
simplex_max_cycles                                           1
simplex_score_converge                                       0.1
simplex_cycle_converge                                       1.0
simplex_trans_step                                           1.0
simplex_rot_step                                             0.1
simplex_tors_step                                            10.0
simplex_anchor_max_iterations                                500
simplex_grow_max_iterations                                  500
simplex_grow_tors_premin_iterations                          0
simplex_random_seed                                          0
simplex_restraint_min                                        yes
simplex_coefficient_restraint                                10.0
atom_model                                                   all
vdw_defn_file                                                {}
flex_defn_file                                               {}
flex_drive_file                                              {}
ligand_outfile_prefix                                        {}_{}_flexdock
write_orientations                                           no
num_scored_conformers                                        20
write_conformations                                          yes
cluster_conformations                                        yes
rank_ligands                                                 yes
        """.format(ligand_path, op.basename(self.sphsel_path), op.splitext(op.basename(self.grid_path))[0],
                   op.splitext(op.basename(self.grid_path))[0], self.amb_file, self.flex1_file, self.flex2_file,
                   self.id, ligand_name)

                f.write(dock_text)

            os.chdir(self.dock_dir)
            cmd = "dock6 -i {} -o {} -v".format(in_name, out_name)
            os.system(cmd)

        if ssbio.utils.is_non_zero_file(ranked_out):
            self.dock_flexible_outfile = out_name
            self.dock_flexible_conformers_result = conformers_out
            self.dock_flexible_scored_result = scored_out
            log.debug('{}: successful docking!'.format(self.dock_flexible_outfile))
        else:
            log.error('{}+{}: empty DOCK6 ranked file, execution error (or ligand failed to dock)'.format(self.id,
                                                                                                          op.basename(ligand_path)))

    def auto_flexdock(self, binding_residues, radius, ligand_path=None, force_rerun=False):
        """Run DOCK6 on a PDB file, given its binding residues and a radius around them.

        Provide a path to a ligand to dock a ligand to it. If no ligand is provided, DOCK6 preparations will be run on
        that structure file.

        Args:
            binding_residues (str): Comma separated string of residues (eg: '144,170,199')
            radius (int, float): Radius around binding residues to dock to
            ligand_path (str): Path to ligand (mol2 format) to dock to protein
            force_rerun (bool): If method should be rerun even if output files exist

        """
        log.debug('\n{}: running DOCK6...\n'
                 '\tBinding residues: {}\n'
                 '\tBinding residues radius: {}\n'
                 '\tLigand to dock: {}\n'.format(self.id, binding_residues, radius, op.basename(ligand_path)))

        self.dockprep(force_rerun=force_rerun)
        self.protein_only_and_noH(force_rerun=force_rerun)
        self.dms_maker(force_rerun=force_rerun)
        self.sphgen(force_rerun=force_rerun)
        self.binding_site_mol2(residues=binding_residues, force_rerun=force_rerun)
        self.sphere_selector_using_residues(radius=radius, force_rerun=force_rerun)
        self.showbox(force_rerun=force_rerun)
        self.grid(force_rerun=force_rerun)

        if ligand_path:
            self.do_dock6_flexible(ligand_path=ligand_path, force_rerun=force_rerun)

def parse_results_mol2(mol2_outpath):
    """Parse a DOCK6 mol2 output file, return a Pandas DataFrame of the results.

    Args:
        mol2_outpath (str): Path to mol2 output file

    Returns:
        DataFrame: Pandas DataFrame of the results

    """
    docked_ligands = pd.DataFrame()

    lines = [line.strip() for line in open(mol2_outpath, 'r')]
    props = {}

    for i, line in enumerate(lines):
        if line.startswith('########## Name:'):
            ligand = line.strip().strip('##########').replace(' ', '').replace('\t', '').split(':')[1]
            line = lines[i + 1]
            props = {}
            props['Ligand'] = ligand
        if line.startswith('##########'):
            splitter = line.strip().strip('##########').replace(' ', '').replace('\t', '').split(':')
            props[splitter[0]] = float(splitter[1])
        if line.startswith('@<TRIPOS>MOLECULE'):
            if props:
                docked_ligands = docked_ligands.append(props, ignore_index=True)

    return docked_ligands

#     def do_dock6_rigid(self, ligand_path, force_rerun=False):
#         ligand_name = os.path.basename(args.ligand).split('.')[0]
#         in_name = op.join(self.dock_dir, "{}_{}_dock.in".format(self.id, ligand_name))
#         out_name = op.join(self.dock_dir, "{}_{}_dock.out".format (self.id, ligand_name))
#
#         with open(in_name, "w") as f:
#             dock_text = """ligand_atom_file                                             {}
# limit_max_ligands                                            no
# skip_molecule                                                no
# read_mol_solvation                                           no
# calculate_rmsd                                               no
# use_database_filter                                          no
# orient_ligand                                                yes
# automated_matching                                           yes
# receptor_site_file                                           {}
# max_orientations                                             1000
# critical_points                                              no
# chemical_matching                                            no
# use_ligand_spheres                                           no
# use_internal_energy                                          yes
# internal_energy_rep_exp                                      12
# flexible_ligand                                              no
# bump_filter                                                  no
# score_molecules                                              yes
# contact_score_primary                                        no
# contact_score_secondary                                      no
# grid_score_primary                                           yes
# grid_score_secondary                                         no
# grid_score_rep_rad_scale                                     1
# grid_score_vdw_scale                                         1
# grid_score_es_scale                                          1
# grid_score_grid_prefix                                       {}
# multigrid_score_secondary                                    no
# dock3.5_score_secondary                                      no
# continuous_score_secondary                                   no
# descriptor_score_secondary                                   no
# gbsa_zou_score_secondary                                     no
# gbsa_hawkins_score_secondary                                 no
# SASA_descriptor_score_secondary                              no
# amber_score_secondary                                        no
# minimize_ligand                                              yes
# simplex_max_iterations                                       1000
# simplex_tors_premin_iterations                               0
# simplex_max_cycles                                           1
# simplex_score_converge                                       0.1
# simplex_cycle_converge                                       1.0
# simplex_trans_step                                           1.0
# simplex_rot_step                                             0.1
# simplex_tors_step                                            10.0
# simplex_random_seed                                          0
# simplex_restraint_min                                        yes
# simplex_coefficient_restraint                                10.0
# atom_model                                                   all
# vdw_defn_file                                                {}
# flex_defn_file                                               {}
# flex_drive_file                                              {}
# ligand_outfile_prefix                                        {}_{}_rigid
# write_orientations                                           yes
# num_scored_conformers                                        20
# rank_ligands                                                 no
#         """.format(ligand_path, self.sphsel_path, self.grid_path.split('.')[0],
#                    self.amb_file, self.flex1_file, self.flex2_file, self.id, ligand_name)
#
#             f.write(dock_text)
#
#         cmd = "dock6 -i {} -o {}".format(in_name, out_name)
#         os.system(cmd)

#     def do_dock6_amberscore(self, ligand_path, force_rerun=False):
#         """INCOMPLETE"""
#         ligand_name = os.path.basename(args.ligand).split('.')[0]
#         in_name = op.join(self.dock_dir, "{}_{}_amberscore.in".format(self.id, ligand_name))
#         out_name = op.join(self.dock_dir, "{}_{}_amberscore.out".format(self.id, ligand_name))
#
#         with open(in_name, "w") as f:
#             dock_text = """ligand_atom_file                                             {}.amber_score.mol2
# limit_max_ligands                                            no
# skip_molecule                                                no
# read_mol_solvation                                           no
# calculate_rmsd                                               no
# use_database_filter                                          no
# orient_ligand                                                no
# use_internal_energy                                          no
# flexible_ligand                                              no
# bump_filter                                                  no
# score_molecules                                              yes
# contact_score_primary                                        no
# contact_score_secondary                                      no
# grid_score_primary                                           no
# grid_score_secondary                                         no
# multigrid_score_primary                                      no
# multigrid_score_secondary                                    no
# dock3.5_score_primary                                        no
# dock3.5_score_secondary                                      no
# continuous_score_primary                                     no
# continuous_score_secondary                                   no
# descriptor_score_primary                                     no
# descriptor_score_secondary                                   no
# gbsa_zou_score_primary                                       no
# gbsa_zou_score_secondary                                     no
# gbsa_hawkins_score_primary                                   no
# gbsa_hawkins_score_secondary                                 no
# SASA_descriptor_score_primary                                no
# SASA_descriptor_score_secondary                              no
# amber_score_primary                                          yes
# amber_score_secondary                                        no
# amber_score_receptor_file_prefix                             {}
# amber_score_movable_region                                   ligand
# amber_score_minimization_rmsgrad                             0.01
# amber_score_before_md_minimization_cycles                    100
# amber_score_md_steps                                         3000
# amber_score_after_md_minimization_cycles                     100
# amber_score_gb_model                                         5
# amber_score_nonbonded_cutoff                                 18.0
# amber_score_temperature                                      300.0
# amber_score_abort_on_unprepped_ligand                        yes
# ligand_outfile_prefix                                        output
# write_orientations                                           no
# num_scored_conformers                                        1
# rank_ligands                                                 no
#         """.format()
#
#             f.write(dock_text)
#
#         cmd = "dock6 -i {} -o {} -v".format(in_name, out_name)
#         os.system(cmd)


# if __name__ == '__main__':
#
#     import glob
#     import argparse
#     import shlex
#
#     # load inputs from command line
#     p = argparse.ArgumentParser(
#             description='Run the DOCK steps on a folder of structure files. To run in the background, execute using nohup: nohup dock.py $BASENAME $NUMFRAMES /path/to/structures/ /path/to/parameters/ --ligand /path/to/ligand.mol2 --cofactors $COFACTORS --residues $RESIDUES > /path/to/logs/$LOGNAME &')
#     p.add_argument('basename', help='base filename that you used to name your files')
#     p.add_argument('numframes', help='total number of frames from your trajectory', type=int)
#     p.add_argument('folder', help='path to folder with your structure files')
#     p.add_argument('params', help='path to folder with parameter files')
#     p.add_argument('--ligand', help='path to file of your ligand that you want to dock')
#     p.add_argument('--cofactors',
#                    help='comma-separated string of cofactors that you want to keep while docking (e.g. SAM,DNC,WAT)')
#     p.add_argument('--residues', help='comma-separated string of the binding residues')
#     p.add_argument('--radius', help='radius around binding residues to dock to (default 9 A)', type=int, default=9)
#     p.add_argument('--redock', help='run DOCK again for the specified ligand, even if docking files exist',
#                    default=False)
#     args = p.parse_args()
#
#     # file paths for docking parameters
#     amb = os.path.join(args.params, 'vdw_AMBER_parm99.defn')
#     f1 = os.path.join(args.params, 'flex.defn')
#     f2 = os.path.join(args.params, 'flex_drive.tbl')
#
#     print(args)
#     # loading current files
#     os.chdir(args.folder)
#     # pdbs = glob.glob('{}-*.pdb'.format(args.basename))
#     current_files = os.listdir(os.getcwd())
#
#     # ligand name
#     if args.ligand:
#         ligandname = os.path.basename(args.ligand)
#
#     # cofactors
#     if args.cofactors:
#         cofactors_list = shlex.split(args.cofactors)
#     else:
#         cofactors_list = []
#
#     print('***************PARAMETERS***************')
#     print('FULL LIST: {0}'.format(vars(args)))
#     if args.ligand:
#         print('LIGAND: {0}'.format(ligandname))
#     if args.cofactors:
#         print('COFACTORS: {0}'.format(cofactors_list))
#     if args.residues:
#         print('BINDING RESIDUES: {0}'.format(args.residues))
#         print('RADIUS: {0}'.format(args.radius))
#
#     counter = 1
#     for frame in range(1, args.numframes + 1):
#         # just a simple counter
#         print(str(counter) + '/' + str(args.numframes))
#         counter += 1
#
#         # file_prefix = '{0}-{1:03d}'.format(args.basename, frame)
#         file_prefix = '{0}'.format(args.basename)
#         print(file_prefix)
#
#         # DOCKPREP
#         # example: 3bwm-440_prep.mol2
#         pdb = '{0}.pdb'.format(file_prefix)
#         prepped_check = '{}_prep.mol2'.format(file_prefix)
#         if prepped_check in current_files:
#             print('***DOCKPREP PREVIOUSLY RUN***')
#             prepped_file = prepped_check
#         else:
#             print('RUNNING: DOCKPREP')
#             prepped_file = dockprep(pdb, file_prefix)
#
#         # ISOLATE RECEPTOR
#         # example: 3bwm-440_receptor.mol2, 3bwm-440_receptor_noH.pdb
#         receptor_check = '{}_receptor.mol2'.format(file_prefix)
#         receptor_noH_check = '{}_receptor_noH.pdb'.format(file_prefix)
#         if receptor_check in current_files and receptor_noH_check in current_files:
#             print('***RECEPTOR FILES PREVIOUSLY GENERATED***')
#             receptor, receptor_noH = receptor_check, receptor_noH_check
#         else:
#             print('RUNNING: ISOLATE RECEPTOR')
#             receptor, receptor_noH = protein_only_and_noH(prepped_file, cofactors_list, file_prefix)
#
#         # DMS
#         # example: 3bwm-440_receptor.dms
#         dms_check = '{}_receptor.dms'.format(file_prefix)
#         if dms_check in current_files:
#             print('***SURFACE PREVIOUSLY GENERATED***')
#             dms = dms_check
#         else:
#             print('RUNNING: DMS')
#             dms = dms_maker(receptor_noH, file_prefix)
#
#         # SPHGEN
#         # example: 3bwm-440_receptor.sph
#         sph_check = '{}_receptor.sph'.format(file_prefix)
#         if sph_check in current_files:
#             print('***SPHERES PREVIOUSLY GENERATED***')
#             sph = sph_check
#         else:
#             print('RUNNING: SPHGEN')
#             sph = sphgen(dms, file_prefix)
#
#         # SPHERE_SELECTOR
#         # first choose binding site and save it as separate .mol2
#         # example: 3BWY-418_binding_residues.mol2
#         binding_site_mol2_file_check = '{}_binding_residues.mol2'.format(file_prefix)
#         if binding_site_mol2_file_check in current_files:
#             print('***BINDING SITE RESIDUES ALREADY DEFINED***')
#             binding_site_mol2_file = binding_site_mol2_file_check
#         else:
#             print('RUNNING: BINDING SITE MOL2')
#             binding_site_mol2_file = binding_site_mol2(receptor_noH, args.residues, file_prefix)
#
#         # then select the spheres based on these binding residues
#         # example: 3bwm-440_selected_spheres_using_binding_residues.sph
#         sel_sph_check = '{}_selected_spheres_using_binding_residues.sph'.format(file_prefix)
#         if sel_sph_check in current_files:
#             print('***SPHERES ALREADY SELECTED***')
#             sel_sph = sel_sph_check
#         else:
#             print('RUNNING: SPHERE_SELECTOR')
#             sel_sph = sphere_selector_using_residues(sph, binding_site_mol2_file, args.radius, file_prefix)
#
#         # SHOWBOX
#         # example: 3bwm-440_box.pdb
#         box_check = '{}_box.pdb'.format(file_prefix)
#         if box_check in current_files:
#             print('***BOX PREVIOUSLY MADE***')
#             box = box_check
#         else:
#             print('RUNNING: SHOWBOX')
#             box = showbox(sel_sph, file_prefix)
#
#         # GRID
#         # example: 3bwm-440_grid.out
#         gr_check = '{}_grid.out'.format(file_prefix)
#         if gr_check in current_files:
#             print('***GRID PREVIOUSLY CALCULATED***')
#             gr = gr_check
#         else:
#             print('RUNNING: GRID')
#             gr = grid(receptor, box, amb, file_prefix)
#
#         # DOCK
#         if args.ligand:
#             dock6_flexible_check = '{}_{}_flexible_scored.mol2'.format((file_prefix, ligandname.split('.')[0]))
#             if dock6_flexible_check in current_files and not args.redock:
#                 print('***DOCK PREVIOUSLY RUN***')
#             else:
#                 print('RUNNING: DOCK')
#                 do_dock6_flexible(args.ligand, sel_sph, gr, amb, f1, f2, file_prefix)
#
#     print('***DOCKING COMPLETE***')



import pytest
from ssbio.pipeline.gempro import GEMPRO

# GEM-PRO with just an ID

# GEM-PRO with ID + root dir

# import shutil
# import os.path as op
# import unittest
# import cobra
# import pandas as pd
# import six
# from ssbio.pipeline.gempro import GEMPRO
# from ssbio.core.genepro import GenePro
# from ssbio.utils import Date
# from ssbio.databases.kegg import KEGGProp
#
# date = Date()
#
# class TestGEMPRO(unittest.TestCase):
#     """Unit tests for the GEMPRO pipeline
#     """
#
#     @classmethod
#     def setUpClass(self):
#         self.GEM_NAME = 'ecoli_test'
#         self.ROOT_DIR = op.join('test_files', 'out')
#         gem_file = op.join('test_files', 'Ec_core_flux1.xml')
#
#         self.my_gempro = GEMPRO(gem_name=self.GEM_NAME, root_dir=self.ROOT_DIR, gem_file_path=gem_file, gem_file_type='sbml')
#         self.base_dir = self.my_gempro.base_dir
#
#         # Remove some genes from this model so testing doesn't take too long
#         remove_these = ['b0118', 'b4025', 'b4153', 'b2925', 'b3919', 'b3738', 'b3732', 'b0726', 'b1602', 'b1101',
#                         'b3236', 'b0728', 'b1603', 'b2926', 'b0432', 'b3735', 'b0474', 'b4090', 'b3731', 'b0767',
#                         'b3737', 'b0724', 'b0008', 'b3403', 'b1779', 'b0727', 'b3956', 'b1676', 'b2935', 'b4232',
#                         'b0430', 'b2914', 'b0723', 'b0722', 'b0729', 'b0720', 'b0116', 'b4152', 'b2415', 'b0429',
#                         'b0621', 'b0904', 'b1241', 'b1276', 'b1380', 'b1416', 'b1479', 'b1612', 'b1702', 'b2417',
#                         'b1852', 'b1854', 'b2276', 'b2277', 'b2278', 'b2279', 'b2280', 'b2281', 'b2282', 'b2283',
#                         'b2284', 'b2285', 'b2286', 'b2287', 'b2288', 'b2297', 'b2463', 'b2465', 'b2587', 'b2975',
#                         'b3117', 'b3386', 'b3528', 'b3603', 'b3739', 'b3951', 'b3952', 'b4122', 'b4395', 'b0721',
#                         'b2029', 'b1136', 'b4015', 'b4014', 'b2976', 'b0114', 'b0115', 'b3916', 'b1723', 'b0755'
#                         ]
#         for x in remove_these:
#             self.my_gempro.genes.remove(self.my_gempro.genes.get_by_id(x))
#
#         # Test the manual adding of any gene ID
#         add_these = ['b0002', 'b0003', 'b0004', 'b2092']
#         self.my_gempro.add_genes_by_id(add_these)
#
#         self.kegg_will_not_map = ['b1417', 'b2092']
#         self.uniprot_will_not_map = ['b1417', 'b2092']
#
#         print(self.base_dir)
#
#     def test_add_genes_by_id(self):
#         add_these = ['b0002', 'b0003', 'b0004', 'b2092']
#         for x in add_these:
#             self.assertTrue(self.my_gempro.genes.has_id(x))
#
#     def test_gene_content(self):
#         # Test that genes were correctly removed from self.genes
#         remaining = ['b2296', 'b3734', 'b3733', 'b3736', 'b0431', 'b2779', 'b4151', 'b4154', 'b1611',  'b2416',
#                      'b0002', 'b0003', 'b0004', 'b1417', 'b2092']
#         six.assertCountEqual(self, [x.id for x in self.my_gempro.genes], remaining)
#
#     def test_prep_folders(self):
#         # Test if folders were made
#         self.assertTrue(op.exists(op.join(self.ROOT_DIR, self.GEM_NAME)))
#         folders = ['data', 'structures', 'sequences']
#         for f in folders:
#             self.assertTrue(op.exists(op.join(self.ROOT_DIR, self.GEM_NAME, f)))
#
#     def test_load_model(self):
#         # Test that we can access the model using .model
#         self.assertIsInstance(self.my_gempro.model, cobra.Model)
#
#         # Test original model attributes
#         self.assertEqual(len(self.my_gempro.model.reactions), 77)
#         self.assertEqual(len(self.my_gempro.model.metabolites), 63)
#         # genes should be removed from model
#         self.assertEqual(len(self.my_gempro.genes), 15)
#
#     def test_gene_info(self):
#         # Test that self.genes is a DictList and contains GenePro objects
#         self.assertIsInstance(self.my_gempro.genes, cobra.DictList)
#         for g in self.my_gempro.genes:
#             self.assertIsInstance(g, GenePro)
#
#     def test_kegg_mapping_and_metadata(self):
#         """Test that KEGG mapping did these things:
#         1. Create the df_kegg_metadata attribute
#         2. Save individual sequence and metadata files per gene in the sequence_files folder
#         3. Save KEGG related info into each GeneInfo object
#         """
#         kegg_organism_code = 'eco'
#         self.my_gempro.kegg_mapping_and_metadata(kegg_organism_code=kegg_organism_code)
#         # print(self.my_gempro.df_kegg_metadata)
#
#         # 1.
#         # Test if attribute is a DataFrame and the number of entries in it is equal to the number of genes
#         self.assertIsInstance(self.my_gempro.df_kegg_metadata, pd.DataFrame)
#         self.assertEqual(len(self.my_gempro.df_kegg_metadata.gene.unique()), len(self.my_gempro.genes))
#
#         ## Tests per gene
#         for gene in self.my_gempro.genes:
#             g = gene.id
#
#             # 2.
#             # Test that gene metadata was downloaded
#             # If the gene did not map to a KEGG ID, a folder will still be made for it
#             # It will still be in the KEGG DF, just with empty entries
#             # It will not have anything downloaded for it
#             gene_structure_folder = op.join(self.my_gempro.sequence_dir, g)
#             self.assertTrue(op.exists(gene_structure_folder))
#             self.assertTrue(g in self.my_gempro.df_kegg_metadata.gene.tolist())
#
#             if g in self.kegg_will_not_map:
#                 self.assertTrue(g in self.my_gempro.missing_kegg_mapping)
#                 self.assertFalse(op.exists(op.join(gene_structure_folder, '{}-{}.faa'.format(kegg_organism_code, g))))
#             else:
#                 self.assertTrue(op.exists(op.join(gene_structure_folder, '{}-{}.faa'.format(kegg_organism_code, g))))
#                 self.assertTrue(op.exists(op.join(gene_structure_folder, '{}-{}.kegg'.format(kegg_organism_code, g))))
#                 # 3.
#                 # Test that the sequence_properties attribute has KEGG information in it
#                 keggs = gene.protein.filter_sequences(KEGGProp)
#                 for k in keggs:
#                     self.assertTrue(k.seq_len > 0)
#
#     def test_uniprot_mapping_and_metadata(self):
#         """Test that UniProt mapping did these things:
#         1. Create the df_uniprot_metadata attribute
#         2. Save individual sequence and metadata files per gene in the sequence_files folder
#         3. Save UniProt related info into each GeneInfo object
#         """
#         self.my_gempro.uniprot_mapping_and_metadata(model_gene_source='ENSEMBLGENOME_ID')
#         # print(self.my_gempro.df_uniprot_metadata)
#
#         # 1.
#         # Test if attribute is a DataFrame and the number of entries in it is equal to the number of genes
#         self.assertIsInstance(self.my_gempro.df_uniprot_metadata, pd.DataFrame)
#         self.assertEqual(len(self.my_gempro.df_uniprot_metadata.gene.unique()), len(self.my_gempro.genes))
#
#         ## Tests per gene
#         for gene in self.my_gempro.genes:
#             g = gene.id
#
#             # 2.
#             # Test that gene metadata was downloaded
#             # If the gene did not map to a UniProt ID, a folder will still be made for it
#             # It will still be in the UniProt DF, just with empty entries
#             # It will not have anything downloaded for it
#             gene_structure_folder = op.join(self.my_gempro.sequence_dir, g)
#             self.assertTrue(op.exists(gene_structure_folder))
#             self.assertTrue(g in self.my_gempro.df_uniprot_metadata.gene.tolist())
#
#             if g in self.uniprot_will_not_map:
#                 self.assertTrue(g in self.my_gempro.missing_uniprot_mapping)
#                 self.assertFalse(op.exists(op.join(gene_structure_folder, '{}.fasta'.format(u))))
#                 self.assertFalse(op.exists(op.join(gene_structure_folder, '{}.txt'.format(u))))
#             else:
#                 uniprots = self.my_gempro.df_uniprot_metadata[self.my_gempro.df_uniprot_metadata.gene == g].uniprot.unique().tolist()
#                 for u in uniprots:
#                     self.assertTrue(op.exists(op.join(gene_structure_folder, '{}.fasta'.format(u))))
#                     self.assertTrue(op.exists(op.join(gene_structure_folder, '{}.txt'.format(u))))
#                     # 3.
#                     # Test that the.annotation['sequence'] attribute has UniProt information in it
#                     u_prop = gene.protein.sequences.get_by_id(u)
#                     self.assertTrue(u_prop.seq_len > 0)
#
#     def test_x_map_uniprot_to_pdb(self):
#         self.my_gempro.uniprot_mapping_and_metadata(model_gene_source='ENSEMBLGENOME_ID')
#         self.my_gempro.set_representative_sequence()
#         self.my_gempro.map_uniprot_to_pdb(seq_ident_cutoff=.9)
#
#         look_at_this_gene = self.my_gempro.genes.get_by_id('b3734').protein
#         self.assertTrue(look_at_this_gene.structures.has_id('3oaa'))
#         look_at_this_structure = look_at_this_gene.structures.get_by_id('3oaa')
#         # TODO: check parsing of taxonomy name
#         # self.assertEqual(536056, look_at_this_structure.taxonomy_name)
#         look_at_this_chain = look_at_this_structure.chains.get_by_id('A')
#         self.assertEqual(513, look_at_this_chain.unp_end)
#         self.assertEqual(1, look_at_this_chain.coverage)
#
#     def test_x_blast_pdb(self):
#         self.my_gempro.uniprot_mapping_and_metadata(model_gene_source='ENSEMBLGENOME_ID')
#         self.my_gempro.set_representative_sequence()
#         self.my_gempro.blast_seqs_to_pdb(seq_ident_cutoff=.8, all_genes=True)
#
#         look_at_this_gene = self.my_gempro.genes.get_by_id('b2296').protein
#         self.assertTrue(look_at_this_gene.structures.has_id('3slc'))
#         look_at_this_structure = look_at_this_gene.structures.get_by_id('3slc')
#         look_at_this_chain = look_at_this_structure.chains.get_by_id('A')
#         self.assertEqual(0.9425, look_at_this_chain.hit_percent_ident)
#
#     @classmethod
#     def tearDownClass(self):
#         shutil.rmtree(self.base_dir)


import os.path as op
import pytest

@pytest.fixture(scope='module')
def test_files():
    return 'test_files'

@pytest.fixture(scope='module')
def test_files_sequences(test_files):
    return op.join(test_files, 'sequences')

@pytest.fixture(scope='module')
def test_files_structures(test_files):
    return op.join(test_files, 'structures')

@pytest.fixture(scope='module')
def test_files_outputs(test_files):
    return op.join(test_files, 'out')


import pytest
import os.path as op
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio.Alphabet import IUPAC
from BCBio import GFF
from ssbio.databases.kegg import KEGGProp

@pytest.fixture(scope='class')
def seq_record_loaded_from_file_example(fasta_path):
    """Original SeqRecord loaded from sequence file"""
    return SeqIO.read(fasta_path, "fasta")

@pytest.fixture(scope='module')
def kegg_id():
    return 'mtu:Rv0417'

@pytest.fixture(scope='module')
def fasta_file():
    return 'mtu-Rv0417.faa'

@pytest.fixture(scope='module')
def txt_file():
    return 'mtu-Rv0417.kegg'

@pytest.fixture(scope='module')
def fasta_path(test_files_sequences, fasta_file):
    return op.join(test_files_sequences, fasta_file)

@pytest.fixture(scope='module')
def txt_path(test_files_sequences, txt_file):
    return op.join(test_files_sequences, txt_file)

@pytest.fixture(scope='class')
def keggprop_with_i(kegg_id):
    return KEGGProp(id=kegg_id,
                    seq=None)

@pytest.fixture(scope='class')
def keggprop_with_i_s_m_f(kegg_id, fasta_path, txt_path):
    return KEGGProp(id=kegg_id,
                    seq=None,
                    fasta_path=fasta_path,
                    txt_path=txt_path)


class TestKEGGPropWithId():

    """Class to test a bare KEGGProp object with just an ID"""

    def test_init(self, keggprop_with_i, kegg_id):
        """Test initializing with just an ID"""
        assert keggprop_with_i.id == kegg_id

        # If just an ID initialized, everything should be empty
        assert keggprop_with_i.seq == None
        assert keggprop_with_i.name == '<unknown name>'
        assert keggprop_with_i.description == '<unknown description>'
        assert len(keggprop_with_i.annotations) == 0
        assert len(keggprop_with_i.letter_annotations) == 0
        assert len(keggprop_with_i.features) == 0

        # Files should not exist and raise errors if accessed
        assert keggprop_with_i.sequence_file == None
        with pytest.raises(IOError):
            keggprop_with_i.sequence_dir
        with pytest.raises(IOError):
            keggprop_with_i.sequence_path
        assert keggprop_with_i.metadata_file == None
        with pytest.raises(IOError):
            keggprop_with_i.metadata_dir
        with pytest.raises(IOError):
            keggprop_with_i.metadata_path
        assert keggprop_with_i.feature_file == None
        with pytest.raises(IOError):
            keggprop_with_i.feature_dir
        with pytest.raises(IOError):
            keggprop_with_i.feature_path

    def test_set_sequence_path(self, keggprop_with_i, fasta_path, fasta_file, test_files_sequences):
        """Test setting the seq attribute with a sequence file"""
        keggprop_with_i.sequence_path = fasta_path

        # Test that file paths are correct
        assert keggprop_with_i.sequence_path == fasta_path
        assert keggprop_with_i.sequence_file == fasta_file
        assert keggprop_with_i.sequence_dir == test_files_sequences

    def test_set_feature_path(self, keggprop_with_i, features_loaded_from_file_example,
                              gff_path, gff_file, test_files_sequences):
        """Test loading a feature file, and that old features are overwritten"""
        # Test that the existing feature set is not the same as the new one to be loaded
        assert len(keggprop_with_i.features) != len(features_loaded_from_file_example)

        keggprop_with_i.feature_path = gff_path

        # Test that file paths are correct
        assert keggprop_with_i.feature_path == gff_path
        assert keggprop_with_i.feature_file == gff_file
        assert keggprop_with_i.feature_dir == test_files_sequences

        # Test that features cannot be changed
        with pytest.raises(ValueError):
            keggprop_with_i.features = ['NOFEATURES']

        # Test that number of features stored is same
        assert len(keggprop_with_i.features) == len(features_loaded_from_file_example)

    def test_set_metadata_path(self, keggprop_with_i, txt_path, txt_file, test_files_sequences,
                               txt_record_loaded_from_file_example):
        keggprop_with_i.metadata_path = txt_path

        # Unset sequence and feature paths
        keggprop_with_i.sequence_path = None
        keggprop_with_i.feature_path = None

        # Test that file paths are correct
        assert keggprop_with_i.metadata_path == txt_path
        assert keggprop_with_i.metadata_file == txt_file
        assert keggprop_with_i.metadata_dir == test_files_sequences

        # Test loaded information
        assert keggprop_with_i.description == txt_record_loaded_from_file_example.description
        assert keggprop_with_i.bigg == None
        for k in ['ecj:JW4347', 'eco:b4384']:
            assert k in keggprop_with_i.kegg
        for r in ['NP_418801.1', 'WP_000224877.1']:
            assert r in keggprop_with_i.refseq
        assert keggprop_with_i.kegg == 'mtu:Rv0417'
        assert keggprop_with_i.gene_name == 'deoD'
        for p in ['1A69', '1ECP', '1K9S', '1OTX', '1OTY', '1OU4', '1OUM', '1OV6', '1OVG',
                    '3ONV', '3OOE', '3OOH', '3OPV', '3UT6', '4TS3', '4TS9', '4TTA', '4TTI',
                    '4TTJ', '5I3C', '5IU6']:
            assert p in keggprop_with_i.pdbs
        for g in ['GO:0004731', 'GO:0005829', 'GO:0006152', 'GO:0006974', 'GO:0016020', 'GO:0019686', 'GO:0042802']:
            assert g in keggprop_with_i.go
        assert keggprop_with_i.pfam == ['PF01048']
        assert keggprop_with_i.ec_number == None  ## TODO: parse
        assert keggprop_with_i.reviewed == False  ## TODO: parse
        for u in ['Q2M5T3', 'P09743']:
            assert u in keggprop_with_i.alt_keggs
        assert keggprop_with_i.taxonomy == 'Escherichia coli (strain K12)'
        assert keggprop_with_i.seq_version == 2
        assert keggprop_with_i.seq_date == '2007-01-23'
        assert keggprop_with_i.entry_version == 106
        assert keggprop_with_i.entry_date == '2017-08-30'

        # Test that features are loaded directly from this metadata file
        assert len(keggprop_with_i.features) == len(txt_record_loaded_from_file_example.features)


class TestKEGGPropWithIdAndFiles():

    """Class to test a bare KEGGProp object with just an ID"""

    def test_init(self, keggprop_with_i_s_m_f, kegg_id,
                  fasta_path, txt_path, gff_path, test_files_sequences,
                  fasta_file, txt_file, gff_file,
                  seq_record_loaded_from_file_example,
                  features_loaded_from_file_example,
                  txt_record_loaded_from_file_example):
        """Test initializing with assigned files"""
        assert keggprop_with_i_s_m_f.id == kegg_id
        assert keggprop_with_i_s_m_f.seq == seq_record_loaded_from_file_example.seq
        assert keggprop_with_i_s_m_f.name == seq_record_loaded_from_file_example.name
        assert keggprop_with_i_s_m_f.description == txt_record_loaded_from_file_example.description
        assert keggprop_with_i_s_m_f.annotations == {}  # No annotations will be loaded from files
        assert keggprop_with_i_s_m_f.letter_annotations == txt_record_loaded_from_file_example.letter_annotations
        assert len(keggprop_with_i_s_m_f.features) == len(features_loaded_from_file_example)

        # Files should exist
        assert keggprop_with_i_s_m_f.sequence_file == fasta_file
        assert keggprop_with_i_s_m_f.sequence_dir == test_files_sequences
        assert keggprop_with_i_s_m_f.sequence_path == fasta_path
        assert keggprop_with_i_s_m_f.metadata_file == txt_file
        assert keggprop_with_i_s_m_f.metadata_dir == test_files_sequences
        assert keggprop_with_i_s_m_f.metadata_path == txt_path
        assert keggprop_with_i_s_m_f.feature_file == gff_file
        assert keggprop_with_i_s_m_f.feature_dir == test_files_sequences
        assert keggprop_with_i_s_m_f.feature_path == gff_path


import logging
import os
import os.path as op
import shutil
import time

import pandas as pd
from ssbio.protein.structure.utils.cleanpdb import CleanPDB
from ssbio.protein.structure.utils.structureio import StructureIO

import ssbio.utils
from ssbio.protein.structure.structprop import StructProp

log = logging.getLogger(__name__)


class ITASSERProp(StructProp):

    """Parse all available information for a local I-TASSER modeling run.

    Initializes a class to collect I-TASSER modeling information and optionally copy results to a new directory.
    SEE: https://zhanglab.ccmb.med.umich.edu/papers/2015_1.pdf for detailed information.

    Args:
        ident (str): ID of I-TASSER modeling run
        original_results_path (str): Path to I-TASSER modeling folder
        coach_results_folder (str): Path to original COACH results
        model_to_use (str): Which I-TASSER model to use. Default is "model1"

    """

    # TODO: parse 1) predicted SS (seq.dat) 2) predicted solvent acc (exp.dat) 3) B-factor (BFP.dat?)
    # Other? Top 10 proteins with similar structure in PDB: $datadir/model1/cofactor/similarpdb_model1.lst

    def __init__(self, ident, original_results_path, coach_results_folder='model1/coach', model_to_use='model1'):

        if not op.exists(original_results_path):
            raise OSError('{}: folder does not exist'.format(original_results_path))

        StructProp.__init__(self, ident, is_experimental=False)

        self._attrs_to_copy = []

        self.original_results_path = original_results_path
        self.model_to_use = model_to_use

        ### MODEL
        self.model_date = None
        original_model_path = op.join(original_results_path, '{}.pdb'.format(model_to_use))
        if not op.exists(original_model_path):
            raise IOError('{}: no homology model generated'.format(original_model_path))
        else:
            self.load_structure_path(structure_path=original_model_path, file_type='pdb')

        ### MODELING RESULTS
        self.difficulty = None
        self.top_template_pdb = None
        self.top_template_chain = None

        # Parse init.dat
        self._init_dat_path = op.join(original_results_path, 'init.dat')
        if op.exists(self._init_dat_path):
            self.update(parse_init_dat(self._init_dat_path))

        # Parse seq.dat
        self._seq_dat_path = op.join(original_results_path, 'seq.dat')
        if op.exists(self._seq_dat_path):
            self._attrs_to_copy.append('_seq_dat_path')
            # TODO: parse seq.dat and store in modeling_results
            self.update(parse_seq_dat(self._seq_dat_path))

        # Parse exp.dat
        self._exp_dat_path = op.join(original_results_path, 'exp.dat')
        if op.exists(self._exp_dat_path):
            self._attrs_to_copy.append('_exp_dat_path')
            # TODO: parse exp.dat and store in modeling_results
            self.update(parse_exp_dat(self._exp_dat_path))

        # Parse BFP.dat
        self._bfp_dat_path = op.join(original_results_path, 'BFP.dat')
        if op.exists(self._bfp_dat_path):
            self._attrs_to_copy.append('_bfp_dat_path')
            # TODO: parse BFP.dat and store in modeling_results
            self.update(parse_bfp_dat(self._bfp_dat_path))

        # Parse cscore
        self.c_score = None
        self.tm_score = None
        self.tm_score_err = None
        self.rmsd = None
        self.rmsd_err = None
        self._cscore_path = op.join(original_results_path, 'cscore')
        if op.exists(self._cscore_path):
            self._attrs_to_copy.append('_cscore_path')
            self.update(parse_cscore(self._cscore_path))


        ### COACH RESULTS
        self.coach_bsites = []
        self.coach_ec = []
        self.coach_go_mf = []
        self.coach_go_bp = []
        self.coach_go_cc = []

        coach_folder = op.join(original_results_path, coach_results_folder)
        self._coach_bsites_inf_path = op.join(coach_folder, 'Bsites.inf')
        self._coach_ec_dat_path = op.join(coach_folder, 'EC.dat')
        self._coach_go_mf_dat_path = op.join(coach_folder, 'GO_MF.dat')
        self._coach_go_bp_dat_path = op.join(coach_folder, 'GO_BP.dat')
        self._coach_go_cc_dat_path = op.join(coach_folder, 'GO_CC.dat')

        if op.isdir(coach_folder):
            # Parse Bsites.inf
            if op.exists(self._coach_bsites_inf_path):
                parsed_bsites = parse_coach_bsites_inf(self._coach_bsites_inf_path)
                if parsed_bsites:
                    self._attrs_to_copy.append('_coach_bsites_inf_path')
                    self.coach_bsites = parsed_bsites

            # Parse EC.dat
            if op.exists(self._coach_ec_dat_path):
                parsed_ec = parse_coach_ec(self._coach_ec_dat_path)
                if parsed_ec:
                    self._attrs_to_copy.append('_coach_ec_dat_path')
                    self.coach_ec = parsed_ec

            # Parse GO_MF.dat
            if op.exists(self._coach_go_mf_dat_path):
                parsed_go_mf = parse_coach_go(self._coach_go_mf_dat_path)
                if parsed_go_mf:
                    self._attrs_to_copy.append('_coach_go_mf_dat_path')
                    self.coach_go_mf = parsed_go_mf

            # Parse GO_BP.dat
            if op.exists(self._coach_go_bp_dat_path):
                parsed_go_bp = parse_coach_go(self._coach_go_bp_dat_path)
                if parsed_go_bp:
                    self._attrs_to_copy.append('_coach_go_bp_dat_path')
                    self.coach_go_bp = parsed_go_bp

            # Parse GO_CC.dat
            if op.exists(self._coach_go_cc_dat_path):
                parsed_go_cc = parse_coach_go(self._coach_go_cc_dat_path)
                if parsed_go_cc:
                    self._attrs_to_copy.append('_coach_go_cc_dat_path')
                    self.coach_go_cc = parsed_go_cc

    def load_structure_path(self, structure_path, file_type='pdb'):
        StructProp.load_structure_path(self, structure_path=structure_path, file_type=file_type)
        # Additionally save model creation date
        self.model_date = time.strftime('%Y-%m-%d', time.gmtime(os.path.getmtime(self.structure_path)))

    def copy_results(self, copy_to_dir, rename_model_to=None, force_rerun=False):
        """Copy the raw information from I-TASSER modeling to a new folder.

        Copies all files in the list _attrs_to_copy.

        Args:
            copy_to_dir (str): Directory to copy the minimal set of results per sequence.
            rename_model_to (str): New file name (without extension)
            force_rerun (bool): If existing models and results should be overwritten.

        """
        # Save path to the structure and copy it if specified
        if not rename_model_to:
            rename_model_to = self.model_to_use

        new_model_path = op.join(copy_to_dir, '{}.pdb'.format(rename_model_to))

        if self.structure_path:
            if ssbio.utils.force_rerun(flag=force_rerun, outfile=new_model_path):
                # Clean and save it
                custom_clean = CleanPDB()
                my_pdb = StructureIO(self.structure_path)
                new_model_path = my_pdb.write_pdb(custom_selection=custom_clean,
                                                  custom_name=rename_model_to,
                                                  out_dir=copy_to_dir,
                                                  force_rerun=force_rerun)

            # Update the structure_path to be the new clean file
            self.load_structure_path(structure_path=new_model_path, file_type='pdb')

            # Other modeling results - store in a new folder
            dest_itasser_dir = op.join(copy_to_dir, '{}_itasser'.format(rename_model_to))
            if not op.exists(dest_itasser_dir):
                os.mkdir(dest_itasser_dir)

            for attr in self._attrs_to_copy:
                old_file_path = getattr(self, attr)
                new_file_path = op.join(dest_itasser_dir, op.basename(old_file_path))
                if ssbio.utils.force_rerun(flag=force_rerun, outfile=new_file_path):
                    shutil.copy2(old_file_path, new_file_path)
                    log.debug('{}: copied from {}'.format(new_file_path, old_file_path))
                else:
                    log.debug('{}: file already exists'.format(new_file_path))
                setattr(self, attr, new_file_path)

    @property
    def df_coach_bsites(self):
        df_cols = ['site_num', 'c_score', 'cluster_size', 'algorithm',
                   'pdb_template_id', 'pdb_template_chain', 'pdb_ligand',
                   'binding_location_coords', 'c_score_method', 'binding_residues',
                   'ligand_cluster_counts']

        bsites_inf_df = pd.DataFrame.from_records(self.coach_bsites, columns=df_cols).drop_duplicates().reset_index(drop=True)

        if bsites_inf_df.empty:
            log.warning('Empty dataframe')
            return bsites_inf_df
        else:
            bsites_inf_df['c_score'] = pd.to_numeric(bsites_inf_df.c_score, errors='coerce')
            bsites_inf_df['cluster_size'] = pd.to_numeric(bsites_inf_df.cluster_size, errors='coerce')
            return ssbio.utils.clean_df(bsites_inf_df)

    @property
    def df_coach_ec(self):
        if self._coach_ec_dat_path:
            return parse_coach_ec_df(self._coach_ec_dat_path)
        else:
            log.warning('Empty dataframe')
            return pd.DataFrame()

    @property
    def df_coach_go(self):
        cols = ['go_id', 'go_term', 'c_score']

        go_all_df = pd.DataFrame()

        for go_list in [self.coach_go_mf, self.coach_go_cc, self.coach_go_bp]:
            go_df = pd.DataFrame.from_records(go_list, columns=cols).drop_duplicates().reset_index(drop=True)
            go_df['c_score'] = pd.to_numeric(go_df.c_score, errors='coerce')

            if go_all_df.empty:
                go_all_df = go_df
            else:
                go_all_df.append(go_df)

        return go_all_df

    def get_dict(self, only_attributes=None, exclude_attributes=None, df_format=False):
        """Summarize the I-TASSER run in a dictionary containing modeling results and top predictions from COACH

        Args:
            only_attributes (str, list): Attributes that should be returned. If not provided, all are returned.
            exclude_attributes (str, list): Attributes that should be excluded.
            df_format (bool): If dictionary values should be formatted for a dataframe
                (everything possible is transformed into strings, int, or float -
                if something can't be transformed it is excluded)

        Returns:
            dict: Dictionary of attributes

        """

        to_exclude = ['coach_bsites', 'coach_ec', 'coach_go_mf', 'coach_go_bp', 'coach_go_cc']
        if not exclude_attributes:
            excluder = to_exclude
        else:
            excluder = ssbio.utils.force_list(exclude_attributes)
            excluder.extend(to_exclude)

        summary_dict = StructProp.get_dict(self, only_attributes=only_attributes,
                                           exclude_attributes=excluder,
                                           df_format=df_format)

        if self.coach_bsites:
            tmp = {'top_bsite_' + k:v for k, v in self.coach_bsites[0].items()}
            summary_dict.update(tmp)

        if self.coach_ec:
            tmp = {'top_ec_' + k: v for k, v in self.coach_ec[0].items()}
            summary_dict.update(tmp)

        if self.coach_go_mf:
            tmp = {'top_go_mf_' + k: v for k, v in self.coach_go_mf[0].items()}
            summary_dict.update(tmp)

        if self.coach_go_bp:
            tmp = {'top_go_bp_' + k: v for k, v in self.coach_go_bp[0].items()}
            summary_dict.update(tmp)

        if self.coach_go_cc:
            tmp = {'top_go_cc_' + k: v for k, v in self.coach_go_cc[0].items()}
            summary_dict.update(tmp)

        return summary_dict


def parse_init_dat(infile):
    """Parse the main init.dat file which contains the modeling results

    The first line of the file init.dat contains stuff like:
    "120 easy  40   8"
    The other lines look like this:
    "     161   11.051   1  1guqA MUSTER"
    and getting the first 10 gives you the top 10 templates used in modeling

    Args:
        infile (stt): Path to init.dat

    Returns:
        dict: Dictionary of parsed information

    """

    # TODO: would be nice to get top 10 templates instead of just the top
    init_dict = {}

    log.debug('{}: reading file...'.format(infile))
    with open(infile, 'r') as f:
        # Get first 2 lines of file
        head = [next(f).strip() for x in range(2)]

    summary = head[0].split()
    difficulty = summary[1]

    top_template_info = head[1].split()
    top_template_pdbchain = top_template_info[3]
    top_template_pdb = top_template_pdbchain[:4]
    top_template_chain = top_template_pdbchain[4:]

    init_dict['difficulty'] = difficulty
    init_dict['top_template_pdb'] = top_template_pdb
    init_dict['top_template_chain'] = top_template_chain

    return init_dict


def parse_seq_dat(infile):
    """Parse the secondary structure predictions in seq.dat

    Args:
        infile (str): Path to seq.dat

    Returns:
        list: List of secondary structure predictions for all residues

    """

    # TODO: parser for seq.dat
    seq_dict = {}
    return seq_dict


def parse_exp_dat(infile):
    """Parse the solvent accessibility predictions in exp.dat

    Args:
        infile (str): Path to exp.dat

    Returns:
        list: List of solvent accessibility predictions for all residues

    """

    # TODO: parser for exp.dat
    exp_dict = {}
    return exp_dict


def parse_bfp_dat(infile):
    """Parse the B-factor predictions in BFP.dat

    Args:
        infile (str): Path to BFP.dat

    Returns:
        list: List of B-factor predictions for all residues

    """

    # TODO: parser for BFP.dat
    bfp_dict = {}
    return bfp_dict


def parse_cscore(infile):
    """Parse the cscore file to return a dictionary of scores.

    Args:
        infile (str): Path to cscore

    Returns:
        dict: Dictionary of scores

    """

    cscore_dict = {}

    with open(infile, 'r') as f:
        for ll in f.readlines():
            # Look for the first line that starts with model1
            if ll.lower().startswith('model1'):
                l = ll.split()

                cscore = l[1]
                tmscore_full = l[2].split('+-')
                tmscore = tmscore_full[0]
                tmscore_err = tmscore_full[1]
                rmsd_full = l[3].split('+-')
                rmsd = rmsd_full[0]
                rmsd_err = rmsd_full[1]

                cscore_dict['c_score'] = float(cscore)
                cscore_dict['tm_score'] = float(tmscore)
                cscore_dict['tm_score_err'] = float(tmscore_err)
                cscore_dict['rmsd'] = float(rmsd)
                cscore_dict['rmsd_err'] = float(rmsd_err)

    return cscore_dict


def parse_coach_bsites_inf(infile):
    """Parse the Bsites.inf output file of COACH and return a list of rank-ordered binding site predictions

    Bsites.inf contains the summary of COACH clustering results after all other prediction algorithms have finished
    For each site (cluster), there are three lines:
    Line 1: site number, c-score of coach prediction, cluster size
    Line 2: algorithm, PDB ID, ligand ID, center of binding site (cartesian coordinates),
        c-score of the algorithm's prediction, binding residues from single template
    Line 3: Statistics of ligands in the cluster

    C-score information
        "In our training data, a prediction with C-score>0.35 has average false positive and false negative rates below
        0.16 and 0.13, respectively." (https://zhanglab.ccmb.med.umich.edu/COACH/COACH.pdf)

    Args:
        infile (str): Path to Bsites.inf

    Returns:
        list: Ranked list of dictionaries, keys defined below
            site_num: cluster which is the consensus binding site
            c_score: confidence score of the cluster prediction
            cluster_size: number of predictions within this cluster
            algorithm: main? algorithm used to make the prediction
            pdb_template_id: PDB ID of the template used to make the prediction
            pdb_template_chain: chain of the PDB which has the ligand
            pdb_ligand: predicted ligand to bind
            binding_location_coords: centroid of the predicted ligand position in the homology model
            c_score_method: confidence score for the main algorithm
            binding_residues: predicted residues to bind the ligand
            ligand_cluster_counts: number of predictions per ligand

    """

    bsites_results = []

    with open(infile) as pp:
        lines = list(filter(None, (line.rstrip() for line in pp)))

    for i in range(len(lines) // 3):
        bsites_site_dict = {}

        line1 = lines[i * 3].split('\t')
        line2 = lines[i * 3 + 1].split('\t')
        line3 = lines[i * 3 + 2]

        bsites_site_dict['site_num'] = line1[0]
        bsites_site_dict['c_score'] = float(line1[1])
        bsites_site_dict['cluster_size'] = line1[2]

        bsites_site_dict['algorithm'] = line2[0]
        bsites_site_dict['pdb_template_id'] = line2[1][:4]
        bsites_site_dict['pdb_template_chain'] = line2[1][4]
        bsites_site_dict['pdb_ligand'] = line2[2]
        bsites_site_dict['binding_location_coords'] = tuple(float(x) for x in line2[3].split())

        # TODO: what's the difference between this c-score and the cluster's c-score?
        # how is the cluster's c-score computed? it's not the average c-score of all methods
        # also why are some COFACTOR c-scores >1?
        # 160411 - seems like the COFACTOR "BS-score" is being reported here, not its c-score...
        tmp_split = line2[4].split(' :')
        bsites_site_dict['c_score_method'] = tmp_split[0]
        bsites_site_dict['binding_residues'] = tmp_split[1]

        bsites_site_dict['ligand_cluster_counts'] = line3

        bsites_results.append(bsites_site_dict)

    return bsites_results


def parse_coach_ec_df(infile):
    """Parse the EC.dat output file of COACH and return a dataframe of results

    EC.dat contains the predicted EC number and active residues.
        The columns are: PDB_ID, TM-score, RMSD, Sequence identity,
        Coverage, Confidence score, EC number, and Active site residues

    Args:
        infile (str): Path to EC.dat

    Returns:
        DataFrame: Pandas DataFrame summarizing EC number predictions

    """

    ec_df = pd.read_table(infile, delim_whitespace=True,
                          names=['pdb_template', 'tm_score', 'rmsd', 'seq_ident', 'seq_coverage',
                                 'c_score', 'ec_number', 'binding_residues'])

    ec_df['pdb_template_id'] = ec_df['pdb_template'].apply(lambda x: x[:4])
    ec_df['pdb_template_chain'] = ec_df['pdb_template'].apply(lambda x: x[4])

    ec_df = ec_df[['pdb_template_id', 'pdb_template_chain', 'tm_score', 'rmsd',
                   'seq_ident', 'seq_coverage', 'c_score', 'ec_number', 'binding_residues']]
    ec_df['c_score'] = pd.to_numeric(ec_df.c_score, errors='coerce')

    return ec_df


def parse_coach_ec(infile):
    """Parse the EC.dat output file of COACH and return a list of rank-ordered EC number predictions

    EC.dat contains the predicted EC number and active residues.
        The columns are: PDB_ID, TM-score, RMSD, Sequence identity,
        Coverage, Confidence score, EC number, and Active site residues

    Args:
        infile (str): Path to EC.dat

    Returns:
         list: Ranked list of dictionaries, keys defined below
            pdb_template_id: PDB ID of the template used to make the prediction
            pdb_template_chain: chain of the PDB which has the ligand
            tm_score: TM-score of the template to the model (similarity score)
            rmsd: RMSD of the template to the model (also a measure of similarity)
            seq_ident: percent sequence identity
            seq_coverage: percent sequence coverage
            c_score: confidence score of the EC prediction
            ec_number: predicted EC number
            binding_residues: predicted residues to bind the ligand

    """
    return list(parse_coach_ec_df(infile).to_dict(orient='index').values())


def parse_coach_go(infile):
    """Parse a GO output file from COACH and return a rank-ordered list of GO term predictions

    The columns in all files are: GO terms, Confidence score, Name of GO terms.
        GO_MF.dat - GO terms in 'molecular function'
        GO_BP.dat - GO terms in 'biological process'
        GO_CC.dat - GO terms in 'cellular component'

    Args:
        infile (str): Path to any COACH GO prediction file

    Returns:
        Pandas DataFrame: Organized dataframe of results, columns defined below
            go_id: GO term ID
            go_term: GO term text
            c_score: confidence score of the GO prediction

    """
    go_list = []

    with open(infile) as go_file:
        for line in go_file.readlines():
            go_dict = {}

            go_split = line.split()
            go_dict['go_id'] = go_split[0]
            go_dict['c_score'] = go_split[1]
            go_dict['go_term'] = ' '.join(go_split[2:])

            go_list.append(go_dict)

    return go_list


import unittest
import os.path as op
import ssbio.protein.structure.properties.fatcat as fatcat


class TestFATCAT(unittest.TestCase):
    """Unit tests for FATCAT"""

    def test_run_fatcat_all_by_all(self):
        OUT_DIR = op.join('test_files', 'out')
        FATCAT_SH = '/home/nathan/software/fatcat/runFATCAT.sh'

        structs = [op.join('test_files', 'structures', '12as-A_clean.pdb'),
                   op.join('test_files', 'structures', '1af6-A_clean.pdb'),
                   op.join('test_files', 'structures', '1a9x-A_clean.pdb')]

        tm_scores = fatcat.run_fatcat_all_by_all(structs, fatcat_sh=FATCAT_SH, outdir=OUT_DIR)



from cobra.core import Gene
from ssbio.core.protein import Protein
import os.path as op
import ssbio.utils
import cobra.io.dict
import logging
log = logging.getLogger(__name__)

class GenePro(Gene):

    """Extends the COBRAPy Gene object to add GEM-PRO annotations in the form of a Protein attribute.

    Additionally adds optional directory paths for files.

    Args:
        id (str): ID for the gene
        name (str): Name of the gene
        functional (bool): If this gene is functional in the linked genome-scale model
        root_dir (str):
        pdb_file_type (str):

    """

    def __init__(self, id, name='', functional=True, root_dir=None, pdb_file_type='mmtf'):
        Gene.__init__(self, id=id, name=name, functional=functional)

        self.pdb_file_type = pdb_file_type

        # Create directories
        self._root_dir = None
        if root_dir:
            self.root_dir = root_dir

        self.protein = Protein(ident=id, root_dir=self.gene_dir, pdb_file_type=self.pdb_file_type)

    @property
    def root_dir(self):
        """Directory where Gene folder is located"""
        return self._root_dir

    @root_dir.setter
    def root_dir(self, path):
        if not path:
            raise ValueError('No path specified')

        if not op.exists(path):
            raise ValueError('{}: folder does not exist'.format(path))

        if self._root_dir:
            log.debug('Changing root directory of Gene "{}" from {} to {}'.format(self.id, self.root_dir, path))

            if not op.exists(op.join(path, self.id)):
                raise IOError('Gene "{}" does not exist in folder {}'.format(self.id, path))

        self._root_dir = path

        ssbio.utils.make_dir(self.gene_dir)

        # Propagate changes to protein
        if hasattr(self, 'protein'):
            self.protein.root_dir = self.gene_dir

    @property
    def gene_dir(self):
        """Gene folder"""
        if self.root_dir:
            return op.join(self.root_dir, self.id)
        else:
            return None

    def reset_protein(self):
        self.protein = Protein(self.id, root_dir=self.gene_dir, pdb_file_type=self.pdb_file_type)

    def __json_encode__(self):
        reqd_attribs = cobra.io.dict._REQUIRED_GENE_ATTRIBUTES
        optn_attribs = cobra.io.dict._OPTIONAL_GENE_ATTRIBUTES
        ordered_optn_attribs = cobra.io.dict._ORDERED_OPTIONAL_GENE_KEYS
        optn_attribs.update({'_root_dir': None})
        ordered_optn_attribs.append('_root_dir')
        new_gene = {key: str(getattr(self, key))
                    for key in reqd_attribs}
        cobra.io.dict._update_optional(self, new_gene, optn_attribs, ordered_optn_attribs)

        # Protein
        new_gene['protein'] = self.protein

        return new_gene

    def __json_decode__(self, **attrs):
        Gene.__init__(self, id=attrs['id'])
        for k, v in attrs.items():
            if k not in ['id']:
                setattr(self, k, v)



import os.path as op
import pandas as pd
import requests
import logging
from copy import copy, deepcopy
from slugify import Slugify

from Bio import SeqIO
from BCBio import GFF
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio.SeqFeature import SeqFeature, FeatureLocation

from ssbio.core.object import Object
import ssbio.utils
import ssbio.databases.pdb
import ssbio.protein.sequence.utils
import ssbio.protein.sequence.utils.fasta
import ssbio.protein.sequence.properties.residues

custom_slugify = Slugify(safe_chars='-_.')
log = logging.getLogger(__name__)


class SeqProp(SeqRecord):

    """Generic class to represent information for a protein sequence.

    Extends the Biopython SeqRecord class. The main functionality added is the ability to set and load directly from
    sequence, metadata, and feature files. Additionally, methods are provided to calculate and store sequence
    properties in the ``annotations`` and ``letter_annotations`` field of a SeqProp. These can then be accessed
    for a range of residue numbers.

    Attributes:
        id (str): Unique identifier for this protein sequence
        seq (Seq): Protein sequence as a Biopython Seq object
        name (str): Optional name for this sequence
        description (str): Optional description for this sequence
        bigg (str, list): BiGG IDs mapped to this sequence
        kegg (str, list): KEGG IDs mapped to this sequence
        refseq (str, list): RefSeq IDs mapped to this sequence
        uniprot (str, list): UniProt IDs mapped to this sequence
        gene_name (str, list): Gene names mapped to this sequence
        pdbs (list): PDB IDs mapped to this sequence
        go (str, list): GO terms mapped to this sequence
        pfam (str, list): PFAMs mapped to this sequence
        ec_number (str, list): EC numbers mapped to this sequence
        sequence_file (str): FASTA file for this sequence
        metadata_file (str): Metadata file (any format) for this sequence
        feature_file (str): GFF file for this sequence
        features (list): List of protein sequence features, which define regions of the protein
        annotations (dict): Annotations of this protein sequence, which summarize global properties
        letter_annotations (RestrictedDict): Residue-level annotations, which describe single residue properties

    Todo:
        - Properly inherit methods from the Object class...

    """

    def __init__(self, seq, id, name='<unknown name>', description='<unknown description>',
                 sequence_path=None, metadata_path=None, feature_path=None):
        """Initialize a SeqProp object with the standard SeqRecord inputs, along with optional paths to files.

        Args:
            seq (str, Seq, SeqRecord): Protein sequence to load
            id (str): Unique identifier for this protein sequence
            name (str): Optional name for this sequence
            description (str): Optional description for this sequence
            sequence_path (str): Absolute or relative path to sequence (FASTA) file
            metadata_path (str): Absolute or relative path to metadata file
            feature_path (str): Absolute or relative path to feature (GFF) file

        """

        # Top level database identifiers
        self.id = id
        self.description = description
        self.notes = {}

        self.bigg = None
        self.kegg = None
        self.refseq = None
        self.uniprot = None
        self.gene_name = None
        self.pdbs = None
        self.go = None
        self.pfam = None
        self.ec_number = None

        # Files, sequences, features
        self._sequence_dir = None
        self.sequence_file = None
        self._metadata_dir = None
        self.metadata_file = None
        self._feature_dir = None
        self.feature_file = None
        self._features = None

        # Stuff from SeqRecord
        self.name = None
        self.description = None
        self.dbxrefs = None
        self.features = None
        self.annotations = None
        self._per_letter_annotations = None

        self._seq = None
        self.seq = seq

        super(SeqProp, self).__init__(seq=self.seq, id=id, name=name, description=description)

        if sequence_path:
            self.sequence_path = sequence_path
        if metadata_path:
            self.metadata_path = metadata_path
        if feature_path:
            self.feature_path = feature_path

    @property
    def seq(self):
        """Seq: Dynamically loaded Seq object from the sequence file"""

        if self.sequence_file:
            file_to_load = copy(self.sequence_path)
            log.debug('{}: reading sequence from sequence file {}'.format(self.id, file_to_load))
            tmp_sr = SeqIO.read(file_to_load, 'fasta')
            return tmp_sr.seq

        else:
            if not self._seq:
                log.debug('{}: no sequence stored in memory'.format(self.id))
            else:
                log.debug('{}: reading sequence from memory'.format(self.id))

            return self._seq

    @seq.setter
    def seq(self, s):
        if not s:
            self._seq = None

        elif self.sequence_file:
            raise ValueError('{}: unable to set sequence, sequence file is associated with this object'.format(self.id))

        elif type(s) == str or type(s) == Seq:
            self._seq = ssbio.protein.sequence.utils.cast_to_seq(obj=s)

        # If a SeqRecord, copy all attributes
        elif type(s) == SeqRecord:
            self._seq = s.seq
            if self.name == '<unknown name>':
                self.name = s.name
            if self.description == '<unknown description>':
                self.description = s.description
            if not self.dbxrefs:
                self.dbxrefs = s.dbxrefs
            if not self.features:
                self.features = s.features
            if not self.annotations:
                self.annotations = s.annotations
            if not self.letter_annotations:
                self.letter_annotations = s.letter_annotations

    @property
    def features(self):
        """list: Get the features stored in memory or in the GFF file"""

        if self.feature_file:
            log.debug('{}: reading features from feature file {}'.format(self.id, self.feature_path))
            with open(self.feature_path) as handle:
                feats = list(GFF.parse(handle))
                if len(feats) > 1:
                    log.warning('Too many sequences in GFF')
                else:
                    return feats[0].features

        else:
            return self._features

    @features.setter
    def features(self, feats):
        if self.feature_file:
            raise ValueError('{}: unable to set features, feature file is associated with this object'.format(self.id))
        elif feats:
            self._features = feats
        else:
            self._features = []

    @property
    def seq_str(self):
        """str: Get the sequence formatted as a string"""

        if not self.seq:
            return None

        return ssbio.protein.sequence.utils.cast_to_str(self.seq)

    @property
    def seq_len(self):
        """int: Get the sequence length"""

        if not self.seq:
            return 0

        return len(self.seq)

    @property
    def num_pdbs(self):
        """int: Report the number of PDB IDs stored in the ``pdbs`` attribute"""

        if not self.pdbs:
            return 0

        else:
            return len(self.pdbs)

    @property
    def sequence_dir(self):
        if not self._sequence_dir:
            raise OSError('No sequence folder set')
        return self._sequence_dir

    @sequence_dir.setter
    def sequence_dir(self, path):
        if path and not op.exists(path):
            raise OSError('{}: folder does not exist'.format(path))

        self._sequence_dir = path

    @property
    def sequence_path(self):
        if not self.sequence_file:
            raise OSError('Sequence file not loaded')

        path = op.join(self.sequence_dir, self.sequence_file)
        if not op.exists(path):
            raise OSError('{}: file does not exist'.format(path))
        return path

    @sequence_path.setter
    def sequence_path(self, fasta_path):
        """Provide pointers to the paths of the FASTA file

        Args:
            fasta_path: Path to FASTA file

        """
        if not fasta_path:
            self.sequence_dir = None
            self.sequence_file = None

        else:
            if not op.exists(fasta_path):
                raise OSError('{}: file does not exist'.format(fasta_path))

            if not op.dirname(fasta_path):
                self.sequence_dir = '.'
            else:
                self.sequence_dir = op.dirname(fasta_path)
            self.sequence_file = op.basename(fasta_path)

            tmp_sr = SeqIO.read(fasta_path, 'fasta')
            if self.name == '<unknown name>':
                self.name = tmp_sr.name
            if self.description == '<unknown description>':
                self.description = tmp_sr.description
            if not self.dbxrefs:
                self.dbxrefs = tmp_sr.dbxrefs
            if not self.features:
                self.features = tmp_sr.features
            if not self.annotations:
                self.annotations = tmp_sr.annotations
            if not self.letter_annotations:
                self.letter_annotations = tmp_sr.letter_annotations

    @property
    def metadata_dir(self):
        if not self._metadata_dir:
            raise OSError('No metadata folder set')
        return self._metadata_dir

    @metadata_dir.setter
    def metadata_dir(self, path):
        if path and not op.exists(path):
            raise OSError('{}: folder does not exist'.format(path))

        self._metadata_dir = path

    @property
    def metadata_path(self):
        if not self.metadata_file:
            raise OSError('Metadata file not loaded')

        path = op.join(self.metadata_dir, self.metadata_file)
        if not op.exists(path):
            raise OSError('{}: file does not exist'.format(path))
        return path

    @metadata_path.setter
    def metadata_path(self, m_path):
        """Provide pointers to the paths of the metadata file

        Args:
            m_path: Path to metadata file

        """
        if not m_path:
            self.metadata_dir = None
            self.metadata_file = None

        else:
            if not op.exists(m_path):
                raise OSError('{}: file does not exist!'.format(m_path))

            if not op.dirname(m_path):
                self.metadata_dir = '.'
            else:
                self.metadata_dir = op.dirname(m_path)
            self.metadata_file = op.basename(m_path)

    @property
    def feature_dir(self):
        if not self._feature_dir:
            raise OSError('No feature folder set')
        return self._feature_dir

    @feature_dir.setter
    def feature_dir(self, path):
        if path and not op.exists(path):
            raise OSError('{}: folder does not exist'.format(path))

        self._feature_dir = path

    @property
    def feature_path(self):
        path = op.join(self.feature_dir, self.feature_file)
        if not op.exists(path):
            raise OSError('{}: file does not exist'.format(path))
        return path

    @feature_path.setter
    def feature_path(self, gff_path):
        """Load a GFF file with information on a single sequence and store features in the ``features`` attribute

        Args:
            gff_path: Path to GFF file.

        """
        if not gff_path:
            self.feature_dir = None
            self.feature_file = None

        else:
            if not op.exists(gff_path):
                raise OSError('{}: file does not exist!'.format(gff_path))

            if not op.dirname(gff_path):
                self.feature_dir = '.'
            else:
                self.feature_dir = op.dirname(gff_path)
            self.feature_file = op.basename(gff_path)

    def __str__(self):
        return Object.__str__(self)

    def __repr__(self):
        return Object.__repr__(self)

    def update(self, newdata, overwrite=False, only_keys=None):
        Object.update(self, newdata=newdata, overwrite=overwrite, only_keys=only_keys)

    def get_dict(self, only_attributes=None, exclude_attributes=None, df_format=False):
        """Get a dictionary of this object's attributes. Optional format for storage in a Pandas DataFrame.

        Args:
            only_attributes (str, list): Attributes that should be returned. If not provided, all are returned.
            exclude_attributes (str, list): Attributes that should be excluded.
            df_format (bool): If dictionary values should be formatted for a dataframe
                (everything possible is transformed into strings, int, or float -
                if something can't be transformed it is excluded)

        Returns:
            dict: Dictionary of attributes

        """

        # Choose attributes to return, return everything in the object if a list is not specified
        if not only_attributes:
            keys = list(self.__dict__.keys())
        else:
            keys = ssbio.utils.force_list(only_attributes)

        # Remove keys you don't want returned
        if exclude_attributes:
            exclude_attributes = ssbio.utils.force_list(exclude_attributes)
            for x in exclude_attributes:
                if x in keys:
                    keys.remove(x)

        # Copy attributes into a new dictionary
        df_dict = {}
        for k, orig_v in self.__dict__.items():
            if k in keys:
                v = deepcopy(orig_v)
                if df_format:
                    if v and not isinstance(v, str) and not isinstance(v, int) and not isinstance(v,
                                                                                                  float) and not isinstance(
                            v, bool):
                        try:
                            df_dict[k] = ssbio.utils.force_string(deepcopy(v))
                        except TypeError:
                            log.warning('{}: excluding attribute from dict, cannot transform into string'.format(k))
                    elif not v and not isinstance(v, int) and not isinstance(v, float):
                        df_dict[k] = None
                    else:
                        df_dict[k] = deepcopy(v)
                else:
                    df_dict[k] = deepcopy(v)
        return df_dict

    def save_dataframes(self, outdir, prefix='df_'):
        Object.save_dataframes(self, outdir=outdir, prefix=prefix)

    def save_pickle(self, outfile, protocol=2):
        Object.save_pickle(self, outfile=outfile, protocol=protocol)

    def __json_encode__(self):
        Object.__json_encode__(self)

    def save_json(self, outfile, compression=False):
        Object.save_json(self, outfile=outfile, compression=compression)

    def equal_to(self, seq_prop):
        """Test if the sequence is equal to another SeqProp's sequence

        Args:
            seq_prop: SeqProp object

        Returns:
            bool: If the sequences are the same

        """
        if not self.seq or not seq_prop or not seq_prop.seq:
            return False

        return self.seq == seq_prop.seq

    def write_fasta_file(self, outfile, force_rerun=False):
        """Write a FASTA file for the protein sequence, ``seq`` will now load directly from this file.

        Args:
            outfile (str): Path to new FASTA file to be written to
            force_rerun (bool): If an existing file should be overwritten

        """
        if ssbio.utils.force_rerun(flag=force_rerun, outfile=outfile):
            SeqIO.write(self, outfile, "fasta")

        # The Seq as it will now be dynamically loaded from the file
        self.sequence_path = outfile

    def write_gff_file(self, outfile, force_rerun=False):
        """Write a GFF file for the protein features, ``features`` will now load directly from this file.

        Args:
            outfile (str): Path to new FASTA file to be written to
            force_rerun (bool): If an existing file should be overwritten

        """
        if ssbio.utils.force_rerun(outfile=outfile, flag=force_rerun):
            with open(outfile, "w") as out_handle:
                GFF.write([self], out_handle)

        self.feature_path = outfile

    def get_biopython_pepstats(self):
        """Run Biopython's built in ProteinAnalysis module.

        Stores statistics in the ``annotations`` attribute.
        """
        if self.seq:
            try:
                pepstats = ssbio.protein.sequence.properties.residues.biopython_protein_analysis(self.seq)
            except KeyError as e:
                log.error('{}: unable to run ProteinAnalysis module, unknown amino acid {}'.format(self.id, e))
                return
            self.annotations.update(pepstats)
        else:
            raise ValueError('{}: no sequence available, unable to run ProteinAnalysis'.format(self.id))

    def get_emboss_pepstats(self):
        """Run the EMBOSS pepstats program on the protein sequence.

        Stores statistics in the ``annotations`` attribute.
        Saves a ``.pepstats`` file of the results where the sequence file is located.
        """
        if not self.sequence_file:
            raise IOError('FASTA file needs to be written for EMBOSS pepstats to be run')
        outfile = ssbio.protein.sequence.properties.residues.emboss_pepstats_on_fasta(infile=self.sequence_path)
        pepstats = ssbio.protein.sequence.properties.residues.emboss_pepstats_parser(outfile)
        self.annotations.update(pepstats)

    def blast_pdb(self, seq_ident_cutoff=0, evalue=0.0001, display_link=False,
                  outdir=None, force_rerun=False):
        """BLAST this sequence to the PDB"""
        if not outdir:
            outdir = self.sequence_dir
            if not outdir:
                raise ValueError('Output directory must be specified')

        if not self.seq_str:
            log.error('{}: no sequence loaded'.format(self.id))
            return None

        try:
            blast_results = ssbio.databases.pdb.blast_pdb(self.seq_str,
                                                          outfile='{}_blast_pdb.xml'.format(custom_slugify(self.id)),
                                                          outdir=outdir,
                                                          force_rerun=force_rerun,
                                                          evalue=evalue,
                                                          seq_ident_cutoff=seq_ident_cutoff,
                                                          link=display_link)
        except requests.ConnectionError as e:
            log.error('{}: BLAST request timed out'.format(self.id))
            print(e)
            return None

        return blast_results

    def get_residue_annotations(self, start_resnum, end_resnum=None):
        """Retrieve letter annotations for a residue or a range of residues

        Args:
            start_resnum (int): Residue number
            end_resnum (int): Optional residue number, specify if a range is desired

        Returns:
            dict: Letter annotations for this residue or residues

        """
        if not end_resnum:
            end_resnum = start_resnum

        # Create a new SeqFeature
        f = SeqFeature(FeatureLocation(start_resnum - 1, end_resnum))

        # Get sequence properties
        return f.extract(self).letter_annotations
